#! /usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import Dict, Union, Any

import numpy as np
import pandas as pd

from dios import DictOfSeries
from saqc.constants import (
    FILTER_ALL,
    UNFLAGGED,
    BAD,
    GOOD,
)

from saqc.core.flags import Flags
from saqc.lib.types import ExternalFlag


ForwardMap = Dict[ExternalFlag, float]
BackwardMap = Dict[float, ExternalFlag]


class TranslationScheme:
    """
    This class provides the basic translation mechanism and should serve as
    a base class for every other translation scheme.

    The general translation is realized through dictionary lookups, altough
    we might need to extend this logic to also allow calls to translation
    functions in the future. Currently at least one `dict` defining the
    'forward' translation  from 'user flags' -> 'internal flags' needs to be
    provided.
    Optionally a second `dict` can be passed to map 'internal flags' -> 'user flags',
    if the latter is not given, this 'backward' translation will inferred as
    the inverse of the 'forward' translation.

    The translation mechanism imposes a few restrictions:
    - The scheme must be well definied, i.e. we need a backward translation for
      every forward translation (each value in `self._forward` needs a key in
      `self._backward`).
    - We need translations for the special flags `saqc.constants.UNFLAGGED` and
      `saqc.constants.BAD`. That implies, that every valid translation scheme
      provides at least one user flag that maps to `BAD` and one that maps to
      `UNFLAGGED`.
    """

    # (internal) threshold flag above which values will be masked
    DFILTER_DEFAULT: float = FILTER_ALL

    # additional arguments and default values the translation scheme accepts
    ARGUMENTS: Dict[str, Any] = {}

    def __init__(self, forward: ForwardMap, backward: BackwardMap):
        """
        Parameters
        ----------
        forward : dict
            A mapping defining the forward translation of scalar flag values

        backward : dict
            A mapping defining the backward translation of scalar flag values

        Note
        ----
        `backward` needs to provide a mapping for the two special flags
        `saqc.constants.UNFLAGGED`, `saqc.constants.BAD`
        """
        if UNFLAGGED not in backward or BAD not in backward:
            raise ValueError(
                f"need translations for the special flags `UNFLAGGED` ({UNFLAGGED})"
                f" and `BAD` ({BAD})"
            )
        self._forward = forward
        self._backward = backward

    @staticmethod
    def _translate(
        flags: Union[Flags, pd.DataFrame, pd.Series],
        trans_map: Union[ForwardMap, BackwardMap],
    ) -> DictOfSeries:
        """
        Translate a given flag data structure to another according to the
        mapping given in `trans_map`

        Parameters
        ----------
        flags : Flags, pd.DataFrame
            The flags to translate

        Returns
        -------
        pd.DataFrame, Flags
        """
        if isinstance(flags, pd.Series):
            flags = flags.to_frame()

        out = DictOfSeries()
        expected = pd.Index(trans_map.values())
        for field in flags.columns:
            out[field] = flags[field].replace(trans_map)
            diff = pd.Index(out[field]).difference(expected)
            if not diff.empty:
                raise ValueError(
                    f"flags were not translated: {diff.drop_duplicates().to_list()}"
                )
        return out

    def __call__(self, flag: ExternalFlag) -> float:
        """
        Translate a scalar 'external flag' to an 'internal flag'

        Parameters
        ----------
        flag : float, int, str
            The external flag to translate

        Returns
        -------
        float
        """
        if flag not in self._forward:
            if flag not in self._backward:
                raise ValueError(f"invalid flag: {flag}")
            return float(flag)
        return self._forward[flag]

    def forward(self, flags: pd.DataFrame) -> Flags:
        """
        Translate from 'external flags' to 'internal flags'

        Parameters
        ----------
        flags : pd.DataFrame
            The external flags to translate

        Returns
        -------
        Flags object
        """
        return Flags(self._translate(flags, self._forward))

    def backward(
        self,
        flags: Flags,
        raw: bool = False,
        attrs: dict | None = None,
        **kwargs,
    ) -> Union[pd.DataFrame, DictOfSeries]:
        """
        Translate from 'internal flags' to 'external flags'

        Parameters
        ----------
        flags : pd.DataFrame
            The external flags to translate

        raw: bool, default False
            if True return data as DictOfSeries, otherwise as pandas DataFrame.

        attrs : dict or None, default None
            global meta information of saqc-object

        Returns
        -------
        pd.DataFrame
        """
        out = self._translate(flags, self._backward)
        out.attrs = attrs or {}
        if not raw:
            out = out.to_df()
        return out


class FloatScheme(TranslationScheme):

    """
    Acts as the default Translator, provides a changeable subset of the
    internal float flags
    """

    _MAP = {
        -np.inf: -np.inf,
        **{k: k for k in np.arange(0, 256, dtype=float)},
    }

    def __init__(self):
        super().__init__(self._MAP, self._MAP)


class SimpleScheme(TranslationScheme):

    """
    Acts as the default Translator, provides a changeable subset of the
    internal float flags
    """

    _FORWARD = {
        "UNFLAGGED": UNFLAGGED,
        "BAD": BAD,
        "OK": GOOD,
    }

    _BACKWARD = {
        UNFLAGGED: "UNFLAGGED",
        np.nan: "UNFLAGGED",
        BAD: "BAD",
        GOOD: "OK",
    }

    def __init__(self):
        super().__init__(forward=self._FORWARD, backward=self._BACKWARD)



import pandas as pd
import warnings


class ItypeWarning(RuntimeWarning):
    pass


class ItypeCastWarning(ItypeWarning):
    pass


class ItypeCastError(RuntimeError):
    pass


class __Itype:
    def __init__(self):
        raise RuntimeError("a Itype class does not allow instances of itself.")


class DtItype(__Itype):
    name = "datetime"
    unique = True
    subtypes = (pd.DatetimeIndex,)
    min_pdindex = pd.DatetimeIndex([])


class IntItype(__Itype):
    name = "integer"
    unique = True
    subtypes = (pd.RangeIndex, pd.Int64Index, pd.UInt64Index, int)
    min_pdindex = pd.Int64Index([])


class FloatItype(__Itype):
    name = "float"
    subtypes = (pd.Float64Index, float)
    unique = True
    min_pdindex = pd.Float64Index([])


# class MultiItype(__Itype):
#     name = "multi"
#     subtypes = (pd.MultiIndex, )
#     unique = ??


class NumItype(__Itype):
    name = "numeric"
    _subitypes = (IntItype, FloatItype)
    subtypes = _subitypes + IntItype.subtypes + FloatItype.subtypes
    unique = False
    min_pdindex = pd.Float64Index([])


class ObjItype(__Itype):
    name = "object"
    unique = False
    _subitypes = (DtItype, IntItype, FloatItype, NumItype, str)
    _otheritypes = (
        pd.CategoricalIndex,
        pd.IntervalIndex,
        pd.PeriodIndex,
        pd.TimedeltaIndex,
        pd.Index,
    )
    subtypes = _subitypes + _otheritypes + DtItype.subtypes + NumItype.subtypes
    min_pdindex = pd.Index([])


def is_itype(obj, itype):
    """Check if obj is a instance of the given itype or its str-alias was given"""

    # todo: iter through itype as it could be a tuple, if called like ``is_itype(o, (t1,t2))``

    # user gave a Itype, like ``DtItype``
    if type(obj) == type and issubclass(obj, itype):
        return True

    # user gave a string, like 'datetime'
    if isinstance(obj, str) and obj == itype.name:
        return True

    return False


def is_itype_subtype(obj, itype):
    """Check if obj is a subclass or a instance of a subclass of the given itype"""

    # user gave a subtype, like ``pd.DatetimeIndex``
    if type(obj) == type and issubclass(obj, itype.subtypes):
        return True

    # user gave a instance of a subtype, like ``pd.Series(..).index``
    if isinstance(obj, itype.subtypes):
        return True

    return False


def is_itype_like(obj, itype):
    """Check if obj is a subclass or a instance of the given itype or any of its subtypes"""
    return is_itype(obj, itype) or is_itype_subtype(obj, itype)


def get_itype(obj):
    """

    Return the according Itype.

    and return the according Itype
    Parameters
    ----------
    obj : {itype string, Itype, pandas.Index, instance of pd.index}
        get the itype fitting for the input

    Examples
    --------
    >>> get_itype("datetime")
    <class 'dios.lib.DtItype'>

    >>> s = pd.Series(index=pd.to_datetime([]))
    >>> get_itype(s.index)
    <class 'dios.lib.DtItype'>

    >>> get_itype(DtItype)
    <class 'dios.lib.DtItype'>

    >>> get_itype(pd.DatetimeIndex)
    <class 'dios.lib.DtItype'>
    """
    if type(obj) == type and issubclass(obj, __Itype):
        return obj

    # check if it is the actual type, not a subtype
    types = [DtItype, IntItype, FloatItype, NumItype, ObjItype]
    for t in types:
        if is_itype(obj, t):
            return t

    for t in types:
        if is_itype_subtype(obj, t):
            return t

    raise ValueError(
        f"{obj} is not a itype, nor any known subtype of a itype, nor a itype string alias"
    )


def _itype_eq(a, b):
    return is_itype(a, b)


def _itype_lt(a, b):
    return is_itype_subtype(a, b)


def _itype_le(a, b):
    return is_itype_like(a, b)


def _find_least_common_itype(iterable_of_series):
    itypes = [NumItype, FloatItype, IntItype, DtItype]
    tlist = [get_itype(s.index) for s in iterable_of_series]
    found = ObjItype
    if tlist:
        for itype in itypes:
            for t in tlist:
                if _itype_le(t, itype):
                    continue
                break
            else:
                found = itype
    return found


################################################################################
# Casting


class CastPolicy:
    force = "force"
    save = "save"
    never = "never"


_CAST_POLICIES = [CastPolicy.force, CastPolicy.save, CastPolicy.never]


def cast_to_itype(series, itype, policy="lossless", err="raise", inplace=False):
    """Cast a series (more explicit the type of the index) to fit the itype of a dios.

    Return the casted series if successful, None otherwise.

    Note:
        This is very basic number-casting, so in most cases, information from
        the old index will be lost after the cast.
    """

    if policy not in _CAST_POLICIES:
        raise ValueError(f"policy={policy}")
    if err not in ["raise", "ignore"]:
        raise ValueError(f"err={err}")
    if not inplace:
        series = series.copy()
    itype = get_itype(itype)

    if series.empty:
        return pd.Series(index=itype.min_pdindex, dtype=series.dtype)

    series.itype = get_itype(series.index)

    # up-cast issn't necessary because a dios with a higher
    # itype always can take lower itypes.
    # series can have dt/int/float/mixed
    # dt    -> dt           -> mixed
    # int   -> int   -> num -> mixed
    # float -> float -> num -> mixed
    # mixed                 -> mixed
    if _itype_le(series.itype, itype):  # a <= b
        return series

    e = f"A series index of type '{type(series.index)}' cannot be casted to Itype '{itype.name}'"

    # cast any -> dt always fail.
    if is_itype(itype, DtItype):
        pass
    else:
        e += f", as forbidden by the cast-policy '{policy}'."

    if policy == CastPolicy.never:
        pass

    elif policy == CastPolicy.force:
        # cast any (dt/float/mixed) -> int
        if is_itype(itype, IntItype):  # a == b
            series.index = pd.RangeIndex(len(series))
            return series
        # cast any (dt/int/mixed) -> float
        # cast any (dt/float/mixed) -> nur
        if is_itype(itype, FloatItype) or is_itype(itype, NumItype):  # a == b or a == c
            series.index = pd.Float64Index(range(len(series)))
            return series

    elif policy == CastPolicy.save:
        # cast int   -> float
        if is_itype(itype, IntItype) and is_itype(
            series.itype, FloatItype
        ):  # a == b and c == d
            series.index = series.index.astype(float)
            return series
        # cast float -> int, maybe if unique
        if is_itype(itype, FloatItype) and is_itype(
            series.itype, IntItype
        ):  # a == b and c == d
            series.index = series.index.astype(int)
            if series.index.is_unique:
                return series
            e = (
                f"The cast with policy {policy} from series index type '{type(series.index)}' to "
                f"itype {itype.name} resulted in a non-unique index."
            )
        # cast mixed -> int/float always fail

    if err == "raise":
        raise ItypeCastError(e)
    else:
        return None


################################################################################
# OPTIONS


class OptsFields:
    """storage class for the keys in `dios_options`

    Use like so: ``dios_options[OptsFields.X] = Opts.Y``.

    See Also
    --------
        Opts: values for the options dict
        dios_options: options dict for module
    """

    mixed_itype_warn_policy = "mixed_itype_policy"
    disp_max_rows = "disp_max_rows "
    disp_min_rows = "disp_min_rows "
    disp_max_cols = "disp_max_vars"
    dios_repr = "dios_repr"


class Opts:
    """storage class for string values for `dios_options`

    Use like so: ``dios_options[OptsFields.X] = Opts.Y``.

    See Also
    --------
        OptsFields: keys for the options dict
        dios_options: options dict for module
    """

    itype_warn = "warn"
    itype_err = "err"
    itype_ignore = "ignore"
    repr_aligned = "aligned"
    repr_indexed = "indexed"


class __DocDummy(dict):
    pass


dios_options = __DocDummy()
dios_options.update(
    **{
        OptsFields.disp_max_rows: 60,
        OptsFields.disp_min_rows: 10,
        OptsFields.disp_max_cols: 10,
        OptsFields.mixed_itype_warn_policy: Opts.itype_warn,
        OptsFields.dios_repr: Opts.repr_indexed,
    }
)

opdoc = f"""Options dictionary for module `dios`.

Use like so: ``dios_options[OptsFields.X] = Opts.Y``.

**Items**:
 * {OptsFields.dios_repr}: {{'indexed', 'aligned'}} default: 'indexed'
    dios default representation if:
     * `indexed`:  show every column with its index
     * `aligned`:  transform to pandas.DataFrame with indexed merged together.
 * {OptsFields.disp_max_rows}  : int
    Maximum numbers of row before truncated to `disp_min_rows`
    in representation of DictOfSeries

 * {OptsFields.disp_min_rows} : int
    min rows to display if `max_rows` is exceeded

 * {OptsFields.disp_max_cols} : int
    Maximum numbers of columns before truncated representation

 * {OptsFields.mixed_itype_warn_policy} : {{'warn', 'err', 'ignore'}}
    How to inform user about mixed Itype

See Also
--------
    OptsFields: keys for the options dict 
    Opts: values for the options dict 

"""
dios_options.__doc__ = opdoc


def _throw_MixedItype_err_or_warn(itype):
    msg = (
        f"Using '{itype.name}' as itype is not recommend. "
        f"As soon as series with different index types are inserted,\n"
        f"indexing and slicing will almost always fail. "
    )

    if dios_options[OptsFields.mixed_itype_warn_policy] in [
        "ignore",
        Opts.itype_ignore,
    ]:
        pass
    elif dios_options[OptsFields.mixed_itype_warn_policy] in [
        "error",
        "err",
        Opts.itype_err,
    ]:
        msg += "Suppress this error by specifying an unitary 'itype' or giving an 'index' to DictOfSeries."
        raise ItypeCastError(msg)
    else:
        msg += "Silence this warning by specifying an unitary 'itype' or giving an 'index' to DictOfSeries."
        warnings.warn(msg, ItypeWarning)
    return


def example_DictOfSeries():
    """Return a example dios.

    Returns
    -------
    DictOfSeries: an example

    Examples
    --------

    >>> from dios import example_DictOfSeries
    >>> di = example_DictOfSeries()
    >>> di
        a |      b |      c |     d |
    ===== | ====== | ====== | ===== |
    0   0 | 2    5 | 4    7 | 6   0 |
    1   7 | 3    6 | 5   17 | 7   1 |
    2  14 | 4    7 | 6   27 | 8   2 |
    3  21 | 5    8 | 7   37 | 9   3 |
    4  28 | 6    9 | 8   47 | 10  4 |
    5  35 | 7   10 | 9   57 | 11  5 |
    6  42 | 8   11 | 10  67 | 12  6 |
    7  49 | 9   12 | 11  77 | 13  7 |
    8  56 | 10  13 | 12  87 | 14  8 |
    9  63 | 11  14 | 13  97 | 15  9 |
    """
    from dios import DictOfSeries

    a = pd.Series(range(0, 70, 7))
    b = pd.Series(range(5, 15, 1))
    c = pd.Series(range(7, 107, 10))
    d = pd.Series(range(0, 10, 1))

    for i, s in enumerate([a, b, c, d]):
        s.index += i * 2

    di = DictOfSeries(dict(a=a, b=b, c=c, d=d))
    return di.copy()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Sequence, Callable, Union

import numpy as np
import pandas as pd
from typing_extensions import Literal

import saqc


class Scores:
    def assignKNNScore(
        self,
        field: Sequence[str],
        target: str,
        n: int = 10,
        func: Callable[[pd.Series], float] = np.sum,
        freq: Union[float, str] = np.inf,
        min_periods: int = 2,
        method: Literal["ball_tree", "kd_tree", "brute", "auto"] = "ball_tree",
        metric: str = "minkowski",
        p: int = 2,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("assignKNNScore", locals())



import pandas as pd
import numpy as np
import time
from .generate_testsets import get_testset, var_prefix

profile_assignment = False

idx = pd.IndexSlice
rows = 0

fir = ["var", "ts", "ass"]
sec = ["df", "a", "b", "dios"]
timingsdf = pd.DataFrame(columns=pd.MultiIndex.from_product([fir, sec]))


def df_timmings(df, t0, t1, v1, v2):
    _t0 = time.time()
    a = df.loc[t0:t1, :]
    _t1 = time.time()
    b = df.loc[:, v1]
    _t2 = time.time()
    if profile_assignment:
        df.loc[t0:t1, v1] = df.loc[t0:t1, v1] * 1111
    _t3 = time.time()

    timingsdf.at[rows, ("ts", "df")] += _t1 - _t0
    timingsdf.at[rows, ("var", "df")] += _t2 - _t1
    timingsdf.at[rows, ("ass", "df")] += _t3 - _t2
    return a, b, df


def a_timings(df, t0, t1, v1, v2):
    _t0 = time.time()
    a = df.loc[t0:t1, :]
    _t1 = time.time()
    b = df.loc[:, v1]
    _t2 = time.time()
    if profile_assignment:
        df.loc[t0:t1, v1] = df.loc[t0:t1, v1] * 1111
    _t3 = time.time()

    timingsdf.at[rows, ("ts", "a")] += _t1 - _t0
    timingsdf.at[rows, ("var", "a")] += _t2 - _t1
    timingsdf.at[rows, ("ass", "a")] += _t3 - _t2
    return a, b, df


def b_timings(df, t0, t1, v1, v2):
    _t0 = time.time()
    a = df.loc[:, t0:t1]
    _t1 = time.time()
    b = df.loc[v1, :]
    _t2 = time.time()
    if profile_assignment:
        df.loc[v1, t0:t1] = df.loc[v1, t0:t1] * 1111
    _t3 = time.time()

    timingsdf.at[rows, ("ts", "b")] += _t1 - _t0
    timingsdf.at[rows, ("var", "b")] += _t2 - _t1
    timingsdf.at[rows, ("ass", "b")] += _t3 - _t2
    return a, b, df


def dios_timings(dios, t0, t1, v1, v2):
    _t0 = time.time()
    a = dios.loc[t0:t1, :]
    _t1 = time.time()
    b = dios.loc[:, v1]
    _t2 = time.time()
    if profile_assignment:
        dios.loc[t0:t1, v1] = dios.loc[t0:t1, v1] * 1111
    _t3 = time.time()

    timingsdf.at[rows, ("ts", "dios")] += _t1 - _t0
    timingsdf.at[rows, ("var", "dios")] += _t2 - _t1
    timingsdf.at[rows, ("ass", "dios")] += _t3 - _t2
    return a, b, dios


def gen_random_timestamps(m, M):
    r = (M - m) * (np.random.randint(10, 90) + np.random.random()) * 0.01
    a, b = m + r, M - r
    return min(a, b), max(a, b)


def find_index_range(obj):
    min_ = None
    max_ = None
    for r in obj:
        m = obj[r].index.min()
        M = obj[r].index.max()
        try:
            min_ = min(min_, m)
            max_ = max(max_, M)
        except TypeError:
            min_ = m
            max_ = M
    return min_, max_


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # do not touch
    rows = 1

    # max increase of of rows
    # 1 = 10 # 2 = 100 # .... # 5 = 100'000
    iterations = 5
    runs = 1
    cols = 10

    profile_assignment = True

    # which to calc and plot
    use_df = False
    use_a = True
    use_b = True
    use_dios = True

    # plot options
    normalize_to_df = True
    plot_xlog = True
    plot_ylog = True

    # ########################

    v1 = "var1"
    v2 = "var2"
    for i in range(iterations):
        rows *= 10

        timingsdf.loc[rows] = (0,) * len(timingsdf.columns)

        df, a, b, dios = get_testset(rows, cols)
        t0, t4 = find_index_range(df)

        if use_df or normalize_to_df:
            for r in range(runs):
                t1, t2 = gen_random_timestamps(t0, t4)
                vr1 = var_prefix + str(np.random.randint(0, cols))
                df_timmings(df, t1, t2, vr1, None)

        if use_a:
            for r in range(runs):
                t1, t2 = gen_random_timestamps(t0, t4)
                vr1 = var_prefix + str(np.random.randint(0, cols))
                a_timings(a, t1, t2, vr1, None)

        if use_b:
            for r in range(runs):
                t1, t2 = gen_random_timestamps(t0, t4)
                vr1 = var_prefix + str(np.random.randint(0, cols))
                b_timings(b, t1, t2, vr1, None)

        if use_dios:
            for r in range(runs):
                t1, t2 = gen_random_timestamps(t0, t4)
                vr1 = var_prefix + str(np.random.randint(0, cols))
                dios_timings(dios, t1, t2, vr1, None)

    # calc the average
    timingsdf /= runs

    pd.set_option("display.max_columns", 100)

    df = timingsdf
    if not profile_assignment:
        df.drop(labels="ass", axis=1, level=0, inplace=True)
    print("timings:")
    print(df)
    df = df.swaplevel(axis=1)
    if normalize_to_df:
        a = df.loc[:, "a"] / df.loc[:, "df"]
        b = df.loc[:, "b"] / df.loc[:, "df"]
        c = df.loc[:, "df"] / df.loc[:, "df"]
        d = df.loc[:, "dios"] / df.loc[:, "df"]
        df.loc[:, "a"] = a.values
        df.loc[:, "b"] = b.values
        df.loc[:, "df"] = c.values
        df.loc[:, "dios"] = d.values
        all = df.copy()
        all.swaplevel(axis=1)
        print("\n\ndiff:")
        print(all)

    a = df.loc[:, ("a", slice(None))]
    b = df.loc[:, ("b", slice(None))]
    dios = df.loc[:, ("dios", slice(None))]
    df = df.loc[:, ("df", slice(None))]

    ax = plt.gca()
    ax.set_title(f"avg of: {runs} runs, columns: {cols}")

    if use_df:
        df.plot(logy=plot_ylog, logx=plot_xlog, linestyle="-", ax=ax)
    if use_a:
        a.plot(logy=plot_ylog, logx=plot_xlog, linestyle="--", ax=ax)
    if use_b:
        b.plot(logy=plot_ylog, logx=plot_xlog, linestyle=":", ax=ax)
    if use_dios:
        dios.plot(logy=plot_ylog, logx=plot_xlog, linestyle="-.", ax=ax)

    plt.show()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Callable, Tuple, Optional, Union
from typing_extensions import Literal
import numpy as np
import pandas as pd
from dios import DictOfSeries

from saqc.constants import *
from saqc.core import register, Flags
from saqc.core.register import _isflagged, processing
from saqc.lib.tools import evalFreqStr, getFreqDelta, filterKwargs
from saqc.lib.ts_operators import shift2Freq, aggregate2Freq
from saqc.funcs.interpolation import interpolateIndex, _SUPPORTED_METHODS
import saqc.funcs.tools as tools


METHOD2ARGS = {
    "inverse_fshift": ("backward", pd.Timedelta),
    "inverse_bshift": ("forward", pd.Timedelta),
    "inverse_nshift": ("nearest", lambda x: pd.Timedelta(x) / 2),
    "inverse_fagg": ("bfill", pd.Timedelta),
    "inverse_bagg": ("ffill", pd.Timedelta),
    "inverse_nagg": ("nearest", lambda x: pd.Timedelta(x) / 2),
    "match": (None, lambda _: "0min"),
}


@processing()
def linear(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    freq: str,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    A method to "regularize" data by interpolating linearly the data at regular timestamp.

    A series of data is considered "regular", if it is sampled regularly (= having uniform sampling rate).

    Interpolated values will get assigned the worst flag within freq-range.

    Note, that the data only gets interpolated at those (regular) timestamps, that have a valid (existing and
    not-na) datapoint preceeding them and one succeeding them within freq range.
    Regular timestamp that do not suffice this condition get nan assigned AND The associated flag will be of value
    ``UNFLAGGED``.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.

    field : str
        The fieldname of the column, holding the data-to-be-regularized.

    flags : saqc.Flags
        Container to store flags of the data.  freq

    freq : str
        An offset string. The frequency of the grid you want to interpolate your data at.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values and shape may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
        Flags values and shape may have changed relatively to the flags input.
    """
    reserved = ["method", "order", "limit", "downgrade"]
    kwargs = filterKwargs(kwargs, reserved)
    return interpolateIndex(data, field, flags, freq, "time", **kwargs)


@processing()
def interpolate(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    freq: str,
    method: _SUPPORTED_METHODS,
    order: int = 1,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    A method to "regularize" data by interpolating the data at regular timestamp.

    A series of data is considered "regular", if it is sampled regularly (= having uniform sampling rate).

    Interpolated values will get assigned the worst flag within freq-range.

    There are available all the interpolations from the pandas.Series.interpolate method and they are called by
    the very same keywords.

    Note, that, to perform a timestamp aware, linear interpolation, you have to pass ``'time'`` as `method`,
    and NOT ``'linear'``.

    Note, that the data only gets interpolated at those (regular) timestamps, that have a valid (existing and
    not-na) datapoint preceeding them and one succeeding them within freq range.
    Regular timestamp that do not suffice this condition get nan assigned AND The associated flag will be of value
    ``UNFLAGGED``.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.

    field : str
        The fieldname of the column, holding the data-to-be-regularized.

    flags : saqc.Flags
        Container to store flags of the data.  freq

    freq : str
        An offset string. The frequency of the grid you want to interpolate your data at.

    method : {"linear", "time", "nearest", "zero", "slinear", "quadratic", "cubic", "spline", "barycentric",
        "polynomial", "krogh", "piecewise_polynomial", "spline", "pchip", "akima"}
        The interpolation method you want to apply.

    order : int, default 1
        If your selected interpolation method can be performed at different *orders* - here you pass the desired
        order.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values and shape may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
        Flags values and shape may have changed relatively to the flags input.
    """
    reserved = ["limit", "downgrade"]
    kwargs = filterKwargs(kwargs, reserved)
    return interpolateIndex(
        data, field, flags, freq, method=method, order=order, **kwargs
    )


# for @processing this would need to handle dfilter
@register(mask=["field"], demask=[], squeeze=[])
def shift(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    freq: str,
    method: Literal["fshift", "bshift", "nshift"] = "nshift",
    freq_check: Optional[Literal["check", "auto"]] = None,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Function to shift data and flags to a regular (equidistant) timestamp grid, according to ``method``.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.

    field : str
        The fieldname of the column, holding the data-to-be-shifted.

    flags : saqc.Flags
        Container to store flags of the data.

    freq : str
        An frequency Offset String that will be interpreted as the sampling rate you want the data to be shifted to.

    method : {'fshift', 'bshift', 'nshift'}, default 'nshift'
        Specifies how misaligned data-points get propagated to a grid timestamp.
        Following choices are available:

        * 'nshift' : every grid point gets assigned the nearest value in its range. (range = +/- 0.5 * `freq`)
        * 'bshift' : every grid point gets assigned its first succeeding value, if one is available in
          the succeeding sampling interval.
        * 'fshift' : every grid point gets assigned its ultimately preceding value, if one is available in
          the preceeding sampling interval.

    freq_check : {None, 'check', 'auto'}, default None

        * ``None`` : do not validate frequency-string passed to `freq`
        * 'check' : estimate frequency and log a warning if estimate miss matches frequency string passed to `freq`,
          or if no uniform sampling rate could be estimated
        * 'auto' : estimate frequency and use estimate. (Ignores `freq` parameter.)

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values and shape may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
        Flags values and shape may have changed relatively to the flags input.
    """
    datcol = data[field]
    if datcol.empty:
        return data, flags

    freq = evalFreqStr(freq, freq_check, datcol.index)

    # do the shift
    datcol = shift2Freq(datcol, method, freq, fill_value=np.nan)

    # do the shift on the history
    kws = dict(method=method, freq=freq)

    history = flags.history[field].apply(
        index=datcol.index,
        func_handle_df=True,
        func=shift2Freq,
        func_kws={**kws, "fill_value": np.nan},
    )

    flags.history[field] = history
    data[field] = datcol
    return data, flags


# for @processing this would need to handle dfilter
@register(mask=["field"], demask=[], squeeze=[])
def resample(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    freq: str,
    func: Callable[[pd.Series], pd.Series] = np.mean,
    method: Literal["fagg", "bagg", "nagg"] = "bagg",
    maxna: Optional[int] = None,
    maxna_group: Optional[int] = None,
    maxna_flags: Optional[int] = None,  # TODO: still a case ??
    maxna_group_flags: Optional[int] = None,
    flag_func: Callable[[pd.Series], float] = max,
    freq_check: Optional[Literal["check", "auto"]] = None,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Function to resample the data.

    The data will be sampled at regular (equidistant) timestamps aka. Grid points.
    Sampling intervals therefore get aggregated with a function, specified by
    'agg_func' parameter and the result gets projected onto the new timestamps with a
    method, specified by "method". The following method (keywords) are available:

    * ``'nagg'``: all values in the range (+/- `freq`/2) of a grid point get
        aggregated with agg_func and assigned to it.
    * ``'bagg'``: all values in a sampling interval get aggregated with agg_func and
        the result gets assigned to the last grid point.
    * ``'fagg'``: all values in a sampling interval get aggregated with agg_func and
        the result gets assigned to the next grid point.


    Note, that. if possible, functions passed to agg_func will get projected
    internally onto pandas.resample methods, wich results in some reasonable
    performance boost - however, for this to work, you should pass functions that
    have the __name__ attribute initialised and the according methods name assigned
    to it. Furthermore, you shouldnt pass numpys nan-functions (``nansum``,
    ``nanmean``,...) because those for example, have ``__name__ == 'nansum'`` and
    they will thus not trigger ``resample.func()``, but the slower ``resample.apply(
    nanfunc)``. Also, internally, no nans get passed to the functions anyway,
    so that there is no point in passing the nan functions.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.

    field : str
        The fieldname of the column, holding the data-to-be-resampled.

    flags : saqc.Flags
        Container to store flags of the data.

    freq : str
        An Offset String, that will be interpreted as the frequency you want to
        resample your data with.

    func : Callable
        The function you want to use for aggregation.

    method: {'fagg', 'bagg', 'nagg'}, default 'bagg'
        Specifies which intervals to be aggregated for a certain timestamp. (preceding,
        succeeding or "surrounding" interval). See description above for more details.

    maxna : {None, int}, default None
        Maximum number NaNs in a resampling interval. If maxna is exceeded, the interval
        is set entirely to NaN.

    maxna_group : {None, int}, default None
        Same as `maxna` but for consecutive NaNs.

    maxna_flags : {None, int}, default None
        Same as `max_invalid`, only applying for the flags. The flag regarded
        as "invalid" value, is the one passed to empty_intervals_flag (
        default=``BAD``). Also this is the flag assigned to invalid/empty intervals.

    maxna_group_flags : {None, int}, default None
        Same as `maxna_flags`, only applying onto flags. The flag regarded as
        "invalid" value, is the one passed to empty_intervals_flag. Also this is the
        flag assigned to invalid/empty intervals.

    flag_func : Callable, default: max
        The function you want to aggregate the flags with. It should be capable of
        operating on the flags dtype (usually ordered categorical).

    freq_check : {None, 'check', 'auto'}, default None

        * ``None``: do not validate frequency-string passed to `freq`
        * ``'check'``: estimate frequency and log a warning if estimate miss matchs
            frequency string passed to 'freq', or if no uniform sampling rate could be
            estimated
        * ``'auto'``: estimate frequency and use estimate. (Ignores `freq` parameter.)

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values and shape may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
        Flags values and shape may have changed relatively to the flags input.
    """

    datcol = data[field]
    freq = evalFreqStr(freq, freq_check, datcol.index)

    datcol = aggregate2Freq(
        datcol,
        method,
        freq,
        func,
        fill_value=np.nan,
        max_invalid_total=maxna,
        max_invalid_consec=maxna_group,
    )

    kws = dict(
        method=method,
        freq=freq,
        agg_func=flag_func,
        fill_value=np.nan,
        max_invalid_total=maxna_flags,
        max_invalid_consec=maxna_group_flags,
    )

    history = flags.history[field].apply(
        index=datcol.index,
        func=aggregate2Freq,
        func_kws=kws,
    )

    data[field] = datcol
    flags.history[field] = history
    return data, flags


def _getChunkBounds(target: pd.Series, flagscol: pd.Series, freq: str):
    chunk_end = target.reindex(flagscol.index, method="bfill", tolerance=freq)
    chunk_start = target.reindex(flagscol.index, method="ffill", tolerance=freq)
    ignore_flags = chunk_end.isna() | chunk_start.isna()
    return ignore_flags


def _inverseInterpolation(
    source: pd.Series, target: pd.Series, freq: str, chunk_bounds
) -> pd.Series:
    source = source.copy()
    if len(chunk_bounds) > 0:
        source[chunk_bounds] = np.nan
    backprojected = source.reindex(target.index, method="bfill", tolerance=freq)
    fwrdprojected = source.reindex(target.index, method="ffill", tolerance=freq)
    return pd.concat([backprojected, fwrdprojected], axis=1).max(axis=1)


def _inverseAggregation(
    source: Union[pd.Series, pd.DataFrame],
    target: Union[pd.Series, pd.DataFrame],
    freq: str,
    method: str,
):
    return source.reindex(target.index, method=method, tolerance=freq)


def _inverseShift(
    source: pd.Series,
    target: pd.Series,
    drop_mask: pd.Series,
    freq: str,
    method: str,
    fill_value,
) -> pd.Series:
    dtype = source.dtype

    target_drops = target[drop_mask]
    target = target[~drop_mask]
    flags_merged = pd.merge_asof(
        source,
        target.index.to_series(name="pre_index"),
        left_index=True,
        right_index=True,
        tolerance=freq,
        direction=method,
    )
    flags_merged.dropna(subset=["pre_index"], inplace=True)
    flags_merged = flags_merged.set_index(["pre_index"]).squeeze()
    target[flags_merged.index] = flags_merged.values

    # reinsert drops
    source = target.reindex(target.index.union(target_drops.index))
    source.loc[target_drops.index] = target_drops.values

    return source.fillna(fill_value).astype(dtype, copy=False)


@register(
    mask=[],
    demask=[],
    squeeze=[],
    handles_target=True,  # target is mandatory in func, so its allowed
)
def concatFlags(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    target: str,
    method: Literal[
        "inverse_fagg",
        "inverse_bagg",
        "inverse_nagg",
        "inverse_fshift",
        "inverse_bshift",
        "inverse_nshift",
        "inverse_interpolation",
    ],
    freq: Optional[str] = None,
    drop: Optional[bool] = False,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The Function appends flags history of ``fields`` to flags history of ``target``.
    Before Appending, columns in ``field`` history are projected onto the target index via ``method``

    method: (field_flag in associated with "field", source_flags associated with "source")

    'inverse_nagg' - all target_flags within the range +/- freq/2 of a field_flag, get assigned this field flags value.
        (if field_flag > target_flag)
    'inverse_bagg' - all target_flags succeeding a field_flag within the range of "freq", get assigned this field flags
        value. (if field_flag > target_flag)
    'inverse_fagg' - all target_flags preceeding a field_flag within the range of "freq", get assigned this field flags
        value. (if field_flag > target_flag)

    'inverse_interpolation' - all target_flags within the range +/- freq of a field_flag, get assigned this source flags value.
        (if field_flag > target_flag)

    'inverse_nshift' - That target_flag within the range +/- freq/2, that is nearest to a field_flag, gets the source
        flags value. (if field_flag > target_flag)
    'inverse_bshift' - That target_flag succeeding a field flag within the range freq, that is nearest to a
        field_flag, gets assigned this field flags value. (if field_flag > target_flag)
    'inverse_nshift' - That target_flag preceeding a field flag within the range freq, that is nearest to a
        field_flag, gets assigned this field flags value. (if field_flag > target_flag)

    'match' - any target_flag with a timestamp matching a field_flags timestamp gets this field_flags value
    (if field_flag > target_flag)

    Note, to undo or backtrack a resampling/shifting/interpolation that has been performed with a certain method,
    you can just pass the associated "inverse" method. Also you should pass the same drop flags keyword.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.

    field : str
        Fieldname of flags history to append.

    flags : saqc.Flags
        Container to store flags of the data.

    target : str
        Field name of flags history to append to.

    method : {'inverse_fagg', 'inverse_bagg', 'inverse_nagg', 'inverse_fshift', 'inverse_bshift', 'inverse_nshift',
             'match'}
        The method used for projection of ``field`` flags onto ``target`` flags. See description above for more details.

    freq : {None, str},default None
        The ``freq`` determines the projection range for the projection method. See above description for more details.
        Defaultly (None), the sampling frequency of ``field`` is used.

    drop : default False
        If set to `True`, the `field` column will be removed after processing

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
        Flags values and shape may have changed relatively to the flags input.
    """
    flagscol = flags[field]
    target_datcol = data[target]
    target_flagscol = flags[target]

    if target_datcol.empty or flagscol.empty:
        return data, flags

    dummy = pd.Series(np.nan, target_flagscol.index, dtype=float)

    if freq is None:
        freq = getFreqDelta(flagscol.index)
        if freq is None and not method == "match":
            raise ValueError(
                'To project irregularly sampled data, either use method="match", or '
                "pass custom projection range to freq parameter."
            )

    if method[-13:] == "interpolation":
        ignore = _getChunkBounds(target_datcol, flagscol, freq)
        func = _inverseInterpolation
        func_kws = dict(freq=freq, chunk_bounds=ignore, target=dummy)

    elif method[-3:] == "agg":
        projection_method = METHOD2ARGS[method][0]
        tolerance = METHOD2ARGS[method][1](freq)
        func = _inverseAggregation
        func_kws = dict(freq=tolerance, method=projection_method, target=dummy)

    elif method[-5:] == "shift":
        drop_mask = target_datcol.isna() | _isflagged(
            target_flagscol, kwargs["dfilter"]
        )
        projection_method = METHOD2ARGS[method][0]
        tolerance = METHOD2ARGS[method][1](freq)
        func = _inverseShift
        kws = dict(
            freq=tolerance, method=projection_method, drop_mask=drop_mask, target=dummy
        )
        func_kws = {**kws, "fill_value": np.nan}

    elif method == "match":
        func = lambda x: x
        func_kws = {}

    else:
        raise ValueError(f"unknown method {method}")

    history = flags.history[field].apply(dummy.index, func, func_kws)
    flags.history[target].append(history)

    if drop:
        data, flags = tools.dropField(data=data, flags=flags, field=field)

    return data, flags



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import pandas as pd
import dios

from saqc.constants import BAD, UNFLAGGED
from saqc.core import initFlagsLike
from saqc.funcs.pattern import flagPatternByDTW
from tests.common import initData


@pytest.fixture
def data():
    return initData(cols=1, start_date="2016-01-01", end_date="2018-12-31", freq="1D")


@pytest.fixture
def field(data):
    return data.columns[0]


def test_flagPattern_dtw():
    data = pd.Series(0, index=pd.date_range(start="2000", end="2001", freq="1d"))
    data.iloc[10:18] = [0, 5, 6, 7, 6, 8, 5, 0]
    pattern = data.iloc[10:18]

    data = dios.DictOfSeries(dict(data=data, pattern_data=pattern))
    flags = initFlagsLike(data, name="data")
    data, flags = flagPatternByDTW(
        data, "data", flags, reference="pattern_data", flag=BAD
    )

    assert all(flags["data"].iloc[10:18] == BAD)
    assert all(flags["data"].iloc[:9] == UNFLAGGED)
    assert all(flags["data"].iloc[18:] == UNFLAGGED)



from .test_setup import *
from pandas.core.dtypes.common import is_scalar


@pytest.mark.parametrize(("idxer", "exp"), [("a", s1), ("c", s3)])
def test__getitem_loc_singleCol(dios_aligned, idxer, exp):
    di = dios_aligned.loc[:, idxer]
    assert isinstance(di, pd.Series)
    assert (di == exp).all()


@pytest.mark.parametrize(("idxer", "exp"), [((1, "a"), s1), ((3, "c"), s3)])
def test__getitem_loc_singleRow_singleCol(dios_aligned, idxer, exp):
    di = dios_aligned.loc[idxer]
    assert is_scalar(di)
    assert di == exp.loc[idxer[0]]


@pytest.mark.parametrize(
    "idxer",
    [
        "x",
        "2",
        1,
        None,
    ],
)
def test__getitem_loc_singleCol_fail(dios_aligned, idxer):
    with pytest.raises((KeyError, TypeError)):
        di = dios_aligned.loc[:, idxer]


# #############################
# __SETITEM__


@pytest.mark.parametrize(
    ("idxer", "exp"),
    [
        (slice(None), [s1 == s1, s2 == s2, s3 == s3, s4 == s4]),
        (C_BLIST, [s1 == s1, s2 != s2, s3 != s3, s4 == s4]),
    ],
)
def test__setitem_loc_singleCol(dios_aligned, idxer, exp):
    di = dios_aligned.copy()
    di.loc[:, idxer] = 99
    for i, c in enumerate(di):
        assert ((di[c] == 99) == exp[i]).all()


VALS = [
    99,
]


@pytest.mark.parametrize("idxerL", R_LOC_INDEXER)
@pytest.mark.parametrize("idxerR", C_LOC_INDEXER)
@pytest.mark.parametrize("val", VALS)
def test__setitem__loc(dios_aligned, idxerL, idxerR, val):
    di = dios_aligned.copy()
    di.loc[idxerL, idxerR] = val
    assert isinstance(di, DictOfSeries)



import pytest

import pandas as pd
import numpy as np
from saqc.lib.rolling import customRoller

n = np.nan


def test_rolling_existence_of_attrs():
    r = pd.DataFrame().rolling(0).validate()
    c = customRoller(pd.DataFrame(), 0, min_periods=0)
    expected = [attr for attr in dir(r) if not attr.startswith("_")]
    result = [attr for attr in dir(c) if not attr.startswith("_")]
    diff = [attr for attr in expected if attr not in result]
    print(diff)
    assert len(diff) == 0


@pytest.fixture()
def data():
    # a series with a symmetrical but not regular index
    left = pd.date_range("2000", freq="1min", periods=3)
    middle = pd.date_range(left[-1], freq="1H", periods=4)
    right = pd.date_range(middle[-1], freq="1min", periods=3)
    s = pd.Series(1, index=pd.Index([*left, *middle[1:-1], *right]))
    return s


@pytest.mark.parametrize(
    "kws, expected",
    [
        (dict(window="0h", closed="neither"), [n, n, n, n, n, n, n, n]),
        (dict(window="1h", closed="neither"), [n, n, n, n, n, n, 1, 2]),
        (dict(window="2h", closed="neither"), [n, n, n, n, 1, 1, 2, 3]),
        (dict(window="3h", closed="neither"), [n, n, n, n, n, 2, 3, 4]),
        (dict(window="4h", closed="neither"), [n, n, n, n, n, n, n, n]),
        # at least #hour NANs at beginning (removed by expanding=False)
        (dict(window="0h", closed="left"), [n, n, n, n, n, n, n, n]),
        (dict(window="1h", closed="left"), [n, n, n, 1, 1, 1, 1, 2]),
        (dict(window="2h", closed="left"), [n, n, n, n, 2, 2, 2, 3]),
        (dict(window="3h", closed="left"), [n, n, n, n, n, 3, 3, 4]),
        (dict(window="4h", closed="left"), [n, n, n, n, n, n, n, n]),
        # at least #hour NANs at beginning (removed by expanding=False)
        (dict(window="0h", closed="right"), [1, 1, 1, 1, 1, 1, 1, 1]),
        (dict(window="1h", closed="right"), [n, n, n, 1, 1, 1, 2, 3]),
        (dict(window="2h", closed="right"), [n, n, n, n, 2, 2, 3, 4]),
        (dict(window="3h", closed="right"), [n, n, n, n, n, 3, 4, 5]),
        (dict(window="4h", closed="right"), [n, n, n, n, n, n, n, n]),
        # at least #hour NANs at beginning (removed by expanding=False)
        (dict(window="0h", closed="both"), [1, 1, 1, 1, 1, 1, 1, 1]),
        (dict(window="1h", closed="both"), [n, n, n, 2, 2, 2, 2, 3]),
        (dict(window="2h", closed="both"), [n, n, n, n, 3, 3, 3, 4]),
        (dict(window="3h", closed="both"), [n, n, n, n, n, 4, 4, 5]),
        (dict(window="4h", closed="both"), [n, n, n, n, n, n, n, n]),
    ],
    ids=lambda x: str(x if not isinstance(x, list) else ""),
)
def test_rolling_expand(data, kws, expected):
    expected = np.array(expected)
    result = customRoller(data, **kws, expand=False).sum()
    result = result.to_numpy()

    print()
    print(
        pd.DataFrame(
            dict(
                orig=data,
                exp=expected,
                res=result,
            ),
            index=data.index,
        )
    )
    assert np.allclose(result, expected, rtol=0, atol=0, equal_nan=True)


# left and right results are swapped
# the expected result is checked inverted, aka x[::-1]
@pytest.mark.parametrize(
    "kws, expected",
    [
        (dict(window="0h", closed="neither"), [n, n, n, n, n, n, n, n]),
        (dict(window="1h", closed="neither"), [n, n, n, n, n, n, 1, 2]),
        (dict(window="2h", closed="neither"), [n, n, n, n, 1, 1, 2, 3]),
        (dict(window="3h", closed="neither"), [n, n, n, n, n, 2, 3, 4]),
        (dict(window="4h", closed="neither"), [n, n, n, n, n, n, n, n]),
        # at least #hour NANs at beginning (removed by expanding=False)
        (dict(window="0h", closed="right"), [n, n, n, n, n, n, n, n]),
        (dict(window="1h", closed="right"), [n, n, n, 1, 1, 1, 1, 2]),
        (dict(window="2h", closed="right"), [n, n, n, n, 2, 2, 2, 3]),
        (dict(window="3h", closed="right"), [n, n, n, n, n, 3, 3, 4]),
        (dict(window="4h", closed="right"), [n, n, n, n, n, n, n, n]),
        # at least #hour NANs at beginning (removed by expanding=False)
        (dict(window="0h", closed="left"), [1, 1, 1, 1, 1, 1, 1, 1]),
        (dict(window="1h", closed="left"), [n, n, n, 1, 1, 1, 2, 3]),
        (dict(window="2h", closed="left"), [n, n, n, n, 2, 2, 3, 4]),
        (dict(window="3h", closed="left"), [n, n, n, n, n, 3, 4, 5]),
        (dict(window="4h", closed="left"), [n, n, n, n, n, n, n, n]),
        # at least #hour NANs at beginning (removed by expanding=False)
        (dict(window="0h", closed="both"), [1, 1, 1, 1, 1, 1, 1, 1]),
        (dict(window="1h", closed="both"), [n, n, n, 2, 2, 2, 2, 3]),
        (dict(window="2h", closed="both"), [n, n, n, n, 3, 3, 3, 4]),
        (dict(window="3h", closed="both"), [n, n, n, n, n, 4, 4, 5]),
        (dict(window="4h", closed="both"), [n, n, n, n, n, n, n, n]),
    ],
    ids=lambda x: str(x if not isinstance(x, list) else ""),
)
def test_rolling_expand_forward(data, kws, expected):
    expected = np.array(expected)[::-1]  # inverted
    result = customRoller(data, **kws, expand=False, forward=True).sum()
    result = result.to_numpy()

    print()
    print(
        pd.DataFrame(
            dict(
                orig=data,
                exp=expected,
                res=result,
            ),
            index=data.index,
        )
    )
    assert np.allclose(result, expected, rtol=0, atol=0, equal_nan=True)


@pytest.mark.parametrize("window", ["0H", "1H", "2H", "3H", "4H"])
@pytest.mark.parametrize("closed", ["both", "neither", "left", "right"])
@pytest.mark.parametrize("center", [False, True], ids=lambda x: f" center={x} ")
@pytest.mark.parametrize("forward", [False, True], ids=lambda x: f" forward={x} ")
@pytest.mark.parametrize(
    "func",
    [
        "sum",
        "count",
        "mean",
        "median",
        "min",
        "max",
        "skew",
        "kurt",
        "cov",
        "corr",
        "sem",
        "var",
        "std",
    ],
)
def test_dtindexer(data, center, closed, window, forward, func):
    print()
    print("forward", forward)
    print("center", center)
    print("closed", closed)
    print("window", window)

    data: pd.Series

    d = data
    cl = closed
    if forward:
        d = data[::-1]
        cl = "right" if closed == "left" else "left" if closed == "right" else closed
    roller = d.rolling(
        window=window,
        closed=cl,
        center=center,
    )

    expected = getattr(roller, func)()
    if forward:
        expected = expected[::-1]

    roller = customRoller(
        obj=data,
        window=window,
        closed=closed,
        center=center,
        forward=forward,
        expand=True,
    )
    result = getattr(roller, func)()

    print()
    print(
        pd.DataFrame(
            dict(
                orig=data,
                exp=expected,
                res=result,
            ),
            index=data.index,
        )
    )

    # pandas bug
    if pd.__version__ < "1.4" and forward:
        result = result[:-1]
        expected = expected[:-1]

    # pandas bug
    # pandas insert a NaN where a valid value should be
    if (
        pd.__version__ < "1.4"
        and forward
        and func in ["sem", "var", "std"]
        and int(window[:-1]) <= 1
    ):
        pytest.skip("fails for pandas < 1.4")

    assert np.allclose(result, expected, rtol=0, atol=0, equal_nan=True)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Tuple, Union, Optional, Callable

import pandas as pd
from typing_extensions import Literal
import numpy as np
from dios import DictOfSeries

from saqc.constants import *
from saqc.core import register, Flags
from saqc.funcs.rolling import _roll
from saqc.funcs.curvefit import _fitPolynomial
from saqc.lib.tools import filterKwargs


@register(mask=["field"], demask=[], squeeze=[])
def calculatePolynomialResidues(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    window: Union[str, int],
    order: int,
    min_periods: Optional[int] = 0,
    **kwargs
) -> Tuple[DictOfSeries, Flags]:
    """
    Fits a polynomial model to the data and calculate the residues.

    The residue  is calculated by fitting a polynomial of degree `order` to a data
    slice of size `window`, that has x at its center.

    Note, that calculating the residues tends to be quite costy, because a function
    fitting is performed for every sample. To improve performance, consider the
    following possibilities:

    In case your data is sampled at an equidistant frequency grid:

    (1) If you know your data to have no significant number of missing values,
    or if you do not want to calculate residues for windows containing missing values
    any way, performance can be increased by setting min_periods=window.

    Note, that the initial and final window/2 values do not get fitted.

    Each residual gets assigned the worst flag present in the interval of
    the original data.

    Parameters
    ----------
    data : dios.DictOfSeries
        The data.

    field : str
        The column, holding the data-to-be-modelled.

    flags : saqc.Flags
        Container to store quality flags to data.

    window : {str, int}
        The size of the window you want to use for fitting. If an integer is passed,
        the size refers to the number of periods for every fitting window. If an
        offset string is passed, the size refers to the total temporal extension. The
        window will be centered around the vaule-to-be-fitted. For regularly sampled
        timeseries the period number will be casted down to an odd number if even.

    order : int
        The degree of the polynomial used for fitting

    min_periods : int or None, default 0
        The minimum number of periods, that has to be available in every values
        fitting surrounding for the polynomial fit to be performed. If there are not
        enough values, np.nan gets assigned. Default (0) results in fitting
        regardless of the number of values present (results in overfitting for too
        sparse intervals). To automatically set the minimum number of periods to the
        number of values in an offset defined window size, pass np.nan.

    Returns
    -------
    data : dios.DictOfSeries
    flags : saqc.Flags
    """
    reserved = ["residues", "set_flags"]
    filterKwargs(kwargs, reserved)
    return _fitPolynomial(
        data=data,
        field=field,
        flags=flags,
        window=window,
        order=order,
        min_periods=min_periods,
        **kwargs,
        # ctrl args
        return_residues=True,
        set_flags=True,
    )


@register(mask=["field"], demask=[], squeeze=[])
def calculateRollingResidues(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    window: Union[str, int],
    func: Callable[[pd.Series], np.ndarray] = np.mean,
    min_periods: Optional[int] = 0,
    center: bool = True,
    **kwargs
) -> Tuple[DictOfSeries, Flags]:
    """
    Calculate the diff of a rolling-window function and the data.

    Note, that the data gets assigned the worst flag present in the original data.

    Parameters
    ----------
    data : dios.DictOfSeries
        The data.
    field : str
        The column to calculate on.
    flags : saqc.Flags
        Container to store quality flags to data.
    window : {int, str}
        The size of the window you want to roll with. If an integer is passed, the size
        refers to the number of periods for every fitting window. If an offset string
        is passed, the size refers to the total temporal extension. For regularly
        sampled timeseries, the period number will be casted down to an odd number if
        ``center=True``.
    func : Callable, default np.mean
        Function to roll with.
    min_periods : int, default 0
        The minimum number of periods to get a valid value
    center : bool, default True
        If True, center the rolling window.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
    """
    reserved = ["return_residues", "set_flags"]
    kwargs = filterKwargs(kwargs, reserved)
    return _roll(
        data=data,
        field=field,
        flags=flags,
        window=window,
        func=func,
        min_periods=min_periods,
        center=center,
        **kwargs,
        # ctrl args
        set_flags=True,
        return_residues=True,
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import numpy as np
import pandas as pd
from typing import Callable

from saqc.constants import BAD
import saqc


class Noise:
    def flagByStatLowPass(
        self,
        field: str,
        func: Callable[[np.array, pd.Series], float],
        window: str,
        thresh: float,
        sub_window: str = None,
        sub_thresh: float = None,
        min_periods: int = None,
        flag: float = BAD,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("flagByStatLowPass", locals())



from dios import *
import pytest
from numpy.random import randint

try:
    from dios.operators import (
        OP_MAP,
        _OP1_MAP,
        _OP2_DIV_MAP,
        _OP2_ARITH_MAP,
        _OP2_BOOL_MAP,
        _OP2_COMP_MAP,
    )
except ModuleNotFoundError:
    from dios.dios.operators import (
        OP_MAP,
        _OP1_MAP,
        _OP2_DIV_MAP,
        _OP2_ARITH_MAP,
        _OP2_BOOL_MAP,
        _OP2_COMP_MAP,
    )

import pandas as pd
import numpy as np

a = pd.Series(range(0, 70, 7), dtype=int)
b = pd.Series(range(5, 15, 1), dtype=int)
c = pd.Series(range(7, 107, 10), dtype=int)
d = pd.Series(range(0, 10, 1), dtype=int)

s1, s2, s3, s4 = a, b, c, d


def df_aligned__():
    return pd.DataFrame(
        dict(
            a=a.copy(),
            b=b.copy(),
            c=c.copy(),
            d=d.copy(),
        )
    )


def dios_aligned__():
    return DictOfSeries(
        dict(
            a=a.copy(),
            b=b.copy(),
            c=c.copy(),
            d=d.copy(),
        )
    )


def dios_unaligned__():
    di = dios_aligned__().copy()
    for i, s in enumerate(di._data):
        s.index = s.index + i * 2
    return di


def df_unaligned__():
    return dios_unaligned__().to_df()


def dios_fuzzy__(nr_cols=None, mincol=0, maxcol=10, itype=None):
    nr_of_cols = nr_cols if nr_cols else randint(mincol, maxcol + 1)

    ns = 10 ** 9
    sec_per_year = 31536000

    ITYPES = [IntItype, FloatItype, DtItype, ObjItype]
    if itype is not None:
        itype = get_itype(itype)
    else:
        itype = ITYPES[randint(0, len(ITYPES))]

    if itype == IntItype:
        f = lambda i: pd.Int64Index(i.astype(int)).unique()
    elif itype == FloatItype:
        f = lambda i: pd.Float64Index(i).unique()
    elif itype == ObjItype:
        f = lambda i: pd.Index(i.astype(int)).unique().astype(str) + "_str"
    else:  # itype == DtItype:
        f = lambda i: pd.to_datetime(i.astype(int) * ns) + pd.Timedelta("30Y")

    di = DictOfSeries(itype=itype)
    for i in range(nr_of_cols):
        start = randint(0, sec_per_year)
        end = start + randint(0, sec_per_year)
        if end > sec_per_year:
            start, end = end - sec_per_year, start

        base = randint(0, 10 + 1)
        exp = randint(1, int(np.log10(end - start + 100)))
        periods = base ** randint(1, exp + 1)
        index = np.linspace(start, end, periods)
        index = f(index)

        arr = randint(0, 10, len(index))
        di[f"c{i}"] = pd.Series(data=arr, dtype=float, index=index)

    return di


@pytest.fixture
def dios_fuzzy():
    return dios_fuzzy__().copy()


@pytest.fixture
def df_aligned():
    return df_aligned__().copy()


@pytest.fixture
def dios_aligned():
    return dios_aligned__().copy()


@pytest.fixture
def df_unaligned():
    return df_unaligned__().copy()


@pytest.fixture
def dios_unaligned():
    return dios_unaligned__().copy()


def diosSeries_eq_dfSeries(
    df_s, di_s, with_msg=False, df_s_name="di_s", di_s_name="df_s"
):
    def fail(msg):
        if with_msg:
            return False, msg
        return False

    assert isinstance(df_s, pd.Series)
    assert isinstance(di_s, pd.Series)

    if df_s.empty and not di_s.empty:
        return fail(
            f"value mismatch: " f"{df_s_name} is missing, but " f"{di_s_name} == {di_s}"
        )

    idiff = di_s.index.difference(df_s.index)
    if not idiff.empty:
        return fail(
            f"index mismatch: "
            f"{di_s_name}.index: {di_s.index.to_list()}, "
            f"{df_s_name}.index: {df_s.index.to_list()}, "
            f"diff: {idiff.to_list()}"
        )

    # compare series
    for i in df_s.index:
        exp = df_s.loc[i]

        # Normally df-nans, from selecting are just not present values
        # in a dios. But if a Nan was inserted in dios on purpose, it is
        # a valid value, so we try to access the value first.
        try:
            val = di_s.loc[i]
        except KeyError:
            # nan in df, missing in dios -> OK
            if np.isnan(exp):
                continue

            # valid val in df, missing in dios -> FAIL
            else:
                return fail(
                    f"value mismatch: "
                    f"{di_s_name}.loc[{i}] == {exp}, but "
                    f"{df_s_name}.loc[{i}] does not exist"
                )

        # inf = np.isinf(exp) and np.isinf(val)
        # sig = np.sign(exp) == np.sign(val)
        # eq_nan = np.isnan(exp) and np.isnan(val)
        # eq_inf = inf and sig
        # eq_vals = exp == val
        # eq = eq_nan or eq_inf or eq_vals
        eq = np.equal(val, exp)
        assert np.isscalar(eq)

        if not eq:
            return fail(
                f"value mismatch: "
                f"{di_s_name}.loc[{i}] == {exp}, but "
                f"{df_s_name}.loc[{i}] == {val}"
            )

    return True, "equal" if with_msg else True


def dios_eq_df(dios, df, dios_dropped_empty_colums=False, with_msg=False):
    def fail(msg):
        if with_msg:
            return False, msg
        return False

    assert isinstance(df, pd.DataFrame)
    assert isinstance(dios, DictOfSeries)

    # check: dios has not more/other cols than df
    notmore = [c for c in dios if c not in df]
    if notmore:
        return fail(
            f"columns mismatch. "
            f"dios: {dios.columns.to_list()}, "
            f"df: {df.columns.to_list()}, "
            f"diff: {notmore}"
        )

    # check: may df has empty cols and dios has no cols
    # at this locations
    miss = [c for c in df if c not in dios]
    if miss:
        if dios_dropped_empty_colums:
            tmp = []
            for c in miss:
                if not df[c].dropna().empty:
                    tmp += [c]
            if tmp:
                return fail(f"columns mismatch: " f"dios missing column(s): {tmp}")
        else:
            return fail(f"columns mismatch: " f"dios missing column(s): {miss}")

    cols = df.columns.intersection(dios.columns)

    for c in cols:
        ok, m = diosSeries_eq_dfSeries(
            df[c], dios[c], di_s_name=f"di[{c}]", df_s_name=f"df[{c}]", with_msg=True
        )
        if not ok:
            return fail(m)

    return True, "equal" if with_msg else True


# 0,1
NICE_SLICE = [slice(None), slice(None, None, 3)]
R_BLIST = [True, False, False, False, True] * 2
C_BLIST = [True, False, False, True]

#              3,4               5       6
R_LOC_SLICE = NICE_SLICE + [slice(2), slice(2, 8)]
R_LOC_LIST = [[1], [3, 4, 5], pd.Series([3, 7])]
#              7            8                  9
R_LOC_BLIST = [R_BLIST, pd.Series(R_BLIST), pd.Series(R_BLIST).values]

#              0,      1,           2,
C_LOC_LIST = [["a"], ["a", "c"], pd.Series(["a", "c"])]
C_LOC_SLICE = NICE_SLICE + [slice("b"), slice("b", "c")]
C_LOC_BLIST = [
    C_BLIST,
    pd.Series(C_BLIST, index=list("abcd")),
    pd.Series(C_BLIST).values,
]

#                 0 1            2           3            4
RC_iLOC_SLICE = NICE_SLICE + [slice(4), slice(-3, -1), slice(-1, 3)]
R_iLOC_LIST = [[7], [6, 8]]
R_iLOC_BLIST = [
    R_BLIST,
    pd.Series(R_BLIST).values,
]  # only list-likes allowed not series-likes
C_iLOC_LIST = [[0], [1, 3]]
C_iLOC_BLIST = [C_BLIST, pd.Series(C_BLIST).values]

MULTIIDXER = [
    df_aligned__() > 9,
    df_aligned__() != df_aligned__(),
    df_aligned__() == df_aligned__(),
    df_aligned__() % 3 == 0,
]
EMPTYIDEXER = [
    [],
    pd.Series(dtype="O"),
]
EMPTY_DF = [pd.DataFrame()]

BASIC_INDEXER = (
    C_LOC_LIST + R_LOC_SLICE + R_LOC_BLIST + MULTIIDXER + EMPTYIDEXER + EMPTY_DF
)
BASIC_INDEXER_FAIL = [
    ["z"],
    ["a", "z"],
    pd.Series(["a", "z"]),
    pd.DataFrame(dict(a=[1, 2, 3])),
]

R_LOC_INDEXER = R_LOC_SLICE + R_LOC_LIST + R_LOC_BLIST + EMPTYIDEXER
C_LOC_INDEXER = C_LOC_SLICE + C_LOC_LIST + C_LOC_BLIST + EMPTYIDEXER

R_iLOC_INDEXER = RC_iLOC_SLICE + R_iLOC_LIST + R_iLOC_BLIST
C_iLOC_INDEXER = RC_iLOC_SLICE + C_iLOC_LIST + C_iLOC_BLIST

O = [[0, 0, 0], [0, 0, 0]]
I = [[1, 1, 1], [1, 1, 1]]
A = [[1, 2, 3], [4, 5, 6]]
B = [[0, 2, 2], [5, 5, 5]]
C = [[3, 2, 0], [1, 0, 3]]
D = [[6, 5, 4], [3, 2, 1]]
DATA_ALIGNED = [O, I, A, B, C, D]

# outer lists could have differnet length, but this would
# make the checks to complicated
EEE = [[], [], []]
O = [[0, 0], [0, 0, 0], [0, 0, 0, 0]]
I = [[1, 1, 1], [1, 1, 1], [1]]
A = [[1], [2, 3], [4, 5, 6]]
B = [[0, 2, 2], [5], [5, 5]]
C = [[3, 2, 0], [1, 0, 3], [0, 0, 0]]
D = [[6], [2], [9]]
DATA_UNALIGNED = [O, I, A, B, C, D, EEE]

# only use if a single matrix is used
ALL = DATA_ALIGNED + DATA_UNALIGNED

OPCOMP = list(_OP2_COMP_MAP)
OPNOCOMP = list(_OP2_ARITH_MAP) + list(_OP2_BOOL_MAP) + list(_OP2_DIV_MAP)
OP2 = OPCOMP + OPNOCOMP
OP1 = list(_OP1_MAP)


def diosFromMatr(mlist):
    l = []
    for m in mlist:
        l.append(DictOfSeries({i: li.copy() for i, li in enumerate(m)}))
    return tuple(l)


@pytest.fixture()
def datetime_series():
    m = randint(2, 1000)
    idx = pd.date_range("2000", "2010", m)
    return pd.Series(range(m), idx)



import time

import pandas as pd
import numpy as np
import datetime as dt
from ..dios import DictOfSeries
import pickle
import os

var_prefix = "var"


def _gen_testset(rowsz, colsz, freq="1min", disalign=True, randstart=True):
    df = pd.DataFrame()
    dos = DictOfSeries()
    start = dt.datetime.strptime("2000-01-01 00:00:00", "%Y-%m-%d %H:%M:%S")
    times = pd.date_range(periods=rowsz, start=start, freq=freq)

    frequ = freq.strip("0123456789")
    freqv = int(freq[: -len(frequ)])

    for i in range(colsz):

        if randstart:
            # generate random startpoint for each series
            r = str(np.random.randint(int(rowsz * 0.05), int(rowsz * 0.6) + 2)) + frequ
            st = start + pd.Timedelta(r)
            times = pd.date_range(periods=rowsz, start=st, freq=freq)

        if disalign:
            if disalign == "random":
                r = np.random.randint(1, i + 2)
            else:
                # total disalign
                r = i
            times += pd.Timedelta(f"{r}ns")

        d = np.random.randint(1, 9, rowsz)
        v = f"var{i}"
        tmp = pd.DataFrame(index=times, data=d, columns=[v])
        df = pd.merge(df, tmp, left_index=True, right_index=True, how="outer")
        dos[v] = tmp.squeeze().copy()

    return df, dos


def get_random_df_and_dios(rowsz, colsz, freq="1min", disalign=True, randstart=True):
    df, _, _, dios, *_ = get_testset(
        rowsz, colsz, freq=freq, disalign=disalign, randstart=randstart
    )
    return df, dios


def get_testset(
    rows,
    cols,
    freq="1s",
    disalign=True,
    randstart=True,
    storagedir=None,
    noresult=False,
):
    if storagedir is None:
        storagedir = os.path.dirname(__file__)
        storagedir = os.path.join(storagedir, "testsets")

    fname = f"set_f{freq}_d{disalign}_r{randstart}_dim{rows}x{cols}.pkl"
    fpath = os.path.join(storagedir, fname)

    # try to get pickled data
    try:
        with open(fpath, "rb") as fh:
            if noresult:
                return
            tup = pickle.load(fh)

            # file/data was present
            return tup
    except (pickle.UnpicklingError, FileNotFoundError):
        pass

    # generate testset(s)
    df, dios = _gen_testset(
        rowsz=rows, colsz=cols, freq=freq, disalign=disalign, randstart=randstart
    )
    df = df.sort_index(axis=0, level=0)
    df_type_a = df.copy().stack(dropna=False).sort_index(axis=0, level=0).copy()
    df_type_b = df.copy().unstack().sort_index(axis=0, level=0).copy()
    tup = df, df_type_a, df_type_b, dios

    # store testsets
    with open(fpath, "wb") as fh:
        pickle.dump(tup, fh)

    if noresult:
        return

    return tup


def gen_all(rrange, crange):
    for r in rrange:
        for c in crange:
            print(r, " x ", c)
            t0 = time.time()
            get_testset(r, c, noresult=True)
            t1 = time.time()
            print(t1 - t0)


if __name__ == "__main__":
    # import time
    #
    # t0 = time.time()
    # for i in range(7):
    #     get_testset(10**i, 10)
    # t1 = time.time()
    # print(t1-t0)

    rr = [10 ** r for r in range(1, 6)]
    c = range(10, 60, 10)
    gen_all(rr, c)



import pytest

from saqc.lib.plotting import makeFig
import pandas as pd
import numpy as np
import saqc
import dios


def test_makeFig():
    # just testing for no errors to occure...
    data = dios.DictOfSeries(
        pd.Series(
            np.linspace(0, 1000, 1000),
            pd.date_range("2000", "2001", periods=1000),
            name="data",
        )
    )
    d_saqc = saqc.SaQC(data)
    d_saqc = (
        d_saqc.flagRange("data", max=500)
        .flagRange("data", max=400)
        .flagRange("data", max=300)
    )

    # not interactive, no storing
    dummy_path = ""

    d_saqc = d_saqc.plot(field="data", path="")
    d_saqc = d_saqc.plot(field="data", path=dummy_path, history="valid", stats=True)
    d_saqc = d_saqc.plot(field="data", path=dummy_path, history="complete")
    d_saqc = d_saqc.plot(
        field="data", path=dummy_path, ax_kwargs={"ylim": "2000-10"}, stats=True
    )



from .test_setup import *


def test_copy_copy_empty(dios_aligned):
    di = dios_aligned
    shallow = di.copy(deep=False)
    deep = di.copy(deep=True)
    empty_w_cols = di.copy_empty(columns=True)
    empty_no_cols = di.copy_empty(columns=False)

    assert di is not shallow
    assert di is not deep
    assert di is not empty_w_cols
    assert di is not empty_no_cols

    for attr in [
        "itype",
        "_itype",
        "_policy",
    ]:
        dios_attr = getattr(di, attr)
        for cop in [shallow, deep, empty_w_cols, empty_no_cols]:
            copy_attr = getattr(cop, attr)
            assert dios_attr == copy_attr

    assert di.columns.equals(shallow.columns)
    assert di.columns.equals(deep.columns)
    assert di.columns.equals(empty_w_cols.columns)
    assert not di.columns.equals(empty_no_cols.columns)

    for i in di:
        assert di._data[i].index is shallow._data[i].index
        assert di._data[i].index is not deep._data[i].index
        di._data[i][0] = 999999
        assert di[i][0] == shallow[i][0]
        assert di[i][0] != deep[i][0]


@pytest.mark.parametrize("left", diosFromMatr(DATA_UNALIGNED))
# we use comp ops just to get some noise in the data
@pytest.mark.parametrize("op", OPCOMP)
def test_all(left, op):
    a = left
    ser = (op(a, a)).all()
    assert isinstance(ser, pd.Series)
    res = [e for e in ser]
    exp = [op(a[col], a[col]) for col in a]
    for i in range(len(res)):
        assert isinstance(exp[i], pd.Series)
        assert (res[i] == exp[i]).all()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from operator import mod
from typing import Tuple

import numpy as np
import pandas as pd
import operator

from dios import DictOfSeries

from saqc.constants import *
from saqc.core import register, Flags
from saqc.core.register import flagging
from saqc.lib.ts_operators import varQC
from saqc.lib.tools import customRoller, getFreqDelta, statPass


@flagging()
def flagConstants(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    thresh: float,
    window: str,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    This functions flags plateaus/series of constant values of length `window` if
    their maximum total change is smaller than thresh.

    Function flags plateaus/series of constant values. Any interval of values y(t),..y(t+n) is flagged, if:

    (1) n > `window`
    (2) `|(y(t + i) - (t + j)|` < `thresh`, for all i,j in [0, 1, ..., n]

    Flag values are (semi-)constant.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        Name of the column, holding the data-to-be-flagged.
    flags : saqc.Flags
        Container to store quality flags to data.
    thresh : float
        Upper bound for the maximum total change of an interval to be flagged constant.
    window : str
        Lower bound for the size of an interval to be flagged constant.
    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The flags object, holding flags and additional informations related to `data`.
        Flags values may have changed, relatively to the flags input.
    """
    d = data[field]

    # min_periods=2 ensures that at least two non-nan values are present
    # in each window and also min() == max() == d[i] is not possible.
    kws = dict(window=window, min_periods=2, expand=False)

    # 1. find starting points of consecutive constant values as a boolean mask
    # 2. fill the whole window with True's
    rolling = customRoller(d, **kws)
    starting_points_mask = rolling.max() - rolling.min() <= thresh
    rolling = customRoller(starting_points_mask, **kws, forward=True)
    # mimic any()
    mask = (rolling.sum() > 0) & d.notna()

    flags[mask, field] = flag
    return data, flags


@flagging()
def flagByVariance(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    window: str = "12h",
    thresh: float = 0.0005,
    maxna: int = None,
    maxna_group: int = None,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Function flags plateaus/series of constant values. Any interval of values y(t),..y(t+n) is flagged, if:

    (1) n > `window`
    (2) variance(y(t),...,y(t+n) < `thresh`

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-flagged.
    flags : saqc.Flags
        Container to store quality flags to data.
    window : str
        Only intervals of minimum size "window" have the chance to get flagged as constant intervals
    thresh : float
        The upper bound, the variance of an interval must not exceed, if the interval wants to be flagged a plateau.
    maxna : int, default None
        Maximum number of NaNs tolerated in an interval. If more NaNs are present, the
        interval is not flagged as plateau.
    maxna_group : int, default None
        Same as `maxna` but for consecutive NaNs.
    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The flags object, holding flags and additional informations related to `data`.
        Flags values may have changed, relatively to the flags input.
    """
    dataseries = data[field]
    delta = getFreqDelta(dataseries.index)
    if not delta:
        raise IndexError("Timeseries irregularly sampled!")

    if maxna is None:
        maxna = np.inf

    if maxna_group is None:
        maxna_group = np.inf

    min_periods = int(np.ceil(pd.Timedelta(window) / pd.Timedelta(delta)))
    window = pd.Timedelta(window)
    to_set = statPass(
        dataseries,
        lambda x: varQC(x, maxna, maxna_group),
        window,
        thresh,
        operator.lt,
        min_periods=min_periods,
    )

    flags[to_set, field] = flag
    return data, flags



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Optional, Union, Callable, Sequence

import numpy as np
import pandas as pd
from typing_extensions import Literal

from saqc.constants import BAD
import saqc


class Outliers:
    def flagByStray(
        self,
        field: str,
        freq: Optional[Union[int, str]] = None,
        min_periods: int = 11,
        iter_start: float = 0.5,
        alpha: float = 0.05,
        flag: float = BAD,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("flagByStray", locals())

    def flagMVScores(
        self,
        field: Sequence[str],
        trafo: Callable[[pd.Series], pd.Series] = lambda x: x,
        alpha: float = 0.05,
        n: int = 10,
        func: Callable[[pd.Series], float] = np.sum,
        iter_start: float = 0.5,
        partition: Optional[Union[int, str]] = None,
        partition_min: int = 11,
        stray_range: Optional[str] = None,
        drop_flagged: bool = False,  # TODO: still a case ?
        thresh: float = 3.5,
        min_periods: int = 1,
        target: str = None,
        flag: float = BAD,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("flagMVScores", locals())

    def flagRaise(
        self,
        field: str,
        thresh: float,
        raise_window: str,
        freq: str,
        average_window: Optional[str] = None,
        raise_factor: float = 2.0,
        slope: Optional[float] = None,
        weight: float = 0.8,
        flag: float = BAD,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("flagRaise", locals())

    def flagMAD(
        self,
        field: str,
        window: str,
        z: float = 3.5,
        flag: float = BAD,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("flagMAD", locals())

    def flagOffset(
        self,
        field: str,
        thresh: float,
        tolerance: float,
        window: Union[int, str],
        thresh_relative: Optional[float] = None,
        flag: float = BAD,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("flagOffset", locals())

    def flagByGrubbs(
        self,
        field: str,
        window: Union[str, int],
        alpha: float = 0.05,
        min_periods: int = 8,
        pedantic: bool = False,
        flag: float = BAD,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("flagByGrubbs", locals())

    def flagRange(
        self,
        field: str,
        min: float = -np.inf,
        max: float = np.inf,
        flag: float = BAD,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("flagRange", locals())

    def flagCrossStatistic(
        self,
        field: Sequence[str],
        thresh: float,
        method: Literal["modZscore", "Zscore"] = "modZscore",
        flag: float = BAD,
        target: str = None,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("flagCrossStatistic", locals())



#! /usr/bin/env python
# -*- coding: utf-8 -*-
__all__ = [
    "T",
    "ArrayLike",
    "PandasLike",
    "DiosLikeT",
    "CurveFitter",
    "ExternalFlag",
    "OptionalNone",
]

from typing import Any, TypeVar, Union, Dict
from typing_extensions import Protocol, Literal
import numpy as np
import pandas as pd
from dios import DictOfSeries

T = TypeVar("T")
ArrayLike = TypeVar("ArrayLike", np.ndarray, pd.Series, pd.DataFrame)
PandasLike = Union[pd.Series, pd.DataFrame, DictOfSeries]
DiosLikeT = Union[DictOfSeries, pd.DataFrame]

ExternalFlag = Union[str, float, int]


# needed for deeper type hinting magic
class CurveFitter(Protocol):
    def __call__(self, data: np.ndarray, *params: float) -> np.ndarray:
        ...


class GenericFunction(Protocol):

    __name__: str
    __globals__: Dict[str, Any]

    def __call__(self, *args: pd.Series) -> PandasLike:
        ...


class OptionalNone:
    pass



#!/usr/bin/env python

from __future__ import annotations

import warnings
import numpy as np
import pandas as pd

from pandas.api.types import is_bool

if pd.__version__ < "1.4":
    import pandas.core.window.indexers as indexers
else:
    import pandas.core.indexers.objects as indexers


class ForwardMixin:
    def __init__(self, *args, forward: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.forward = forward

    def get_window_bounds(
        self,
        num_values: int = 0,
        min_periods: int | None = None,
        center: bool | None = None,
        closed: str | None = None,
    ) -> tuple[np.ndarray, np.ndarray]:

        if closed is None:
            closed = "right"

        if self.forward:

            # this is only set with variable window indexer
            if self.index_array is not None:  # noqa
                self.index_array = self.index_array[::-1]  # noqa

            # closed 'both' and 'neither' works the same
            # on forward and backward windows by definition
            if closed == "left":
                closed = "right"
            elif closed == "right":
                closed = "left"

        start, end = super().get_window_bounds(  # noqa
            num_values, min_periods, center, closed
        )

        if self.forward:
            start, end = end, start
            start = num_values - start[::-1]
            end = num_values - end[::-1]

        return start, end


CustomFixedIndexer = type(
    "CustomFixedIndexer", (ForwardMixin, indexers.FixedWindowIndexer), {}
)

CustomVariableIndexer = type(
    "CustomVariableIndexer", (ForwardMixin, indexers.VariableWindowIndexer), {}
)


class AttrWrapper(object):
    """
    This wraps a attribute like `rolling.closed` to `customRoller.closed`.
    """

    def __init__(self, name):
        self.name = name

    def __get__(self, instance, owner):
        return getattr(instance._roller, self.name)

    def __set__(self, instance, value):
        setattr(instance._roller, self.name, value)


class CustomRoller:
    def __init__(
        self,
        obj: pd.DataFrame | pd.Series,
        window: int | str | pd.Timedelta,
        min_periods: int | None = None,  # aka minimum non-nan values
        center: bool | None = False,
        win_type: str | None = None,
        on: str | None = None,
        axis: int | str = 0,
        closed: str | None = None,
        forward: bool = False,
        expand=None,
    ):
        """
        A custom rolling implementation, using pandas as base.

        Parameters
        ----------
        obj : pd.Series (or pd.DataFrame)
            The object to roll over. DataFrame is currently still experimental.

        window : int or offset
            Size of the moving window. This is the number of observations used for
            calculating the statistic. Each window will be a fixed size. If its an
            offset then this will be the time period of each window. Each window will
            be a variable sized based on the observations included in the
            time-period. This is only valid for datetimelike indexes.

        min_periods : int, default None
            Minimum number of observations in window required to have a value (
            otherwise result is NA). For a window that is specified by an offset,
            min_periods will default to 1. Otherwise, min_periods will default to the
            size of the window.

        center : bool, default False
            Set the labels at the center of the window. Also works for offset-based
            windows (in contrary to pandas).

        win_type : str, default None
            Not implemented. Raise NotImplementedError if not None.

        on : str, optional
            For a DataFrame, a datetime-like column or MultiIndex level on which to
            calculate the rolling window, rather than the DataFrame’s index. Provided
            integer column is ignored and excluded from result since an integer index
            is not used to calculate the rolling window.

        axis : int or str, default 0

        closed : str, default None
            Make the interval closed on the `right`, `left`, `both` or `neither`
            endpoints. For offset-based windows, with ``forward=False`` it defaults
            to `right`, for ``forward=True`` it defaults to `left`.
            For fixed windows, defaults to ‘both’ always.

        forward : bool, default False
            By default a window is 'looking' backwards (in time). If True the window
            is looking forward in time.

        expand : bool, default True
            If True the window expands/shrink up to its final window size while
            shifted in the data or shifted out respectively. For (normal)
            backward-windows it only expands at the left border, for forward-windows
            it shrinks on the right border and for centered windows both apply.

            Also bear in mind that even if this is True, an many as `min_periods` values are necessary to get a
            valid value, see there for more info.

        Returns
        -------
        a Window or Rolling sub-classed for the particular operation


        Notes
        -----
        If for some reason the start and end numeric indices of the window are needed, one can call
        `start, end = customRoller(obj, window).window.get_window_bounds(num_values, min_periods)`,
        which return two np.arrays, that are holding the start and end indices. Fill at least all
        parameter which are shown in the example.

        See Also
        --------
        pandas.Series.rolling
        pandas.DataFrame.rolling
        """

        if not is_bool(forward):
            raise ValueError("forward must be a boolean")

        # only relevant for datetime-like windows
        if expand is None:
            if min_periods is None:
                expand = False
                warnings.warn(
                    "`expand` defaults to False, if min_periods is None. The result "
                    "will differ from pandas rolling implementation. To fallback to "
                    "pandas rolling, use `expand=True`, to silence this warning and "
                    "use our rolling use `expand=False` or specify `min_periods`."
                )
            else:
                expand = True

        if not is_bool(expand):
            raise TypeError(f"expand must be bool or None not {type(expand).__name__}")

        # ours
        self._forward = forward
        self._expand = expand

        # dummy roller.
        # This lets pandas do all the checks.
        verified = obj.rolling(
            window=window,
            min_periods=min_periods,
            center=center,
            win_type=win_type,
            on=on,
            axis=axis,
            closed=closed,
        )

        self._dtlike_window = verified._win_freq_i8 is not None

        # these roller attributes are fixed by us and not
        # get looked up from self._roller in __getattr__
        # because they might not be correct when passing
        # a custom indexer to rolling
        self.window = verified.window
        self._win_freq_i8 = verified._win_freq_i8
        self._index_array = verified._index_array

        if self._dtlike_window:
            self.window_indexer = CustomVariableIndexer(
                index_array=verified._index_array,
                window_size=verified._win_freq_i8,
                center=verified.center,
                forward=self._forward,
            )
        else:
            self.window_indexer = CustomFixedIndexer(
                window_size=verified.window,
                forward=self._forward,
            )

        # set the default
        if self._forward and closed is None:
            closed = "left"

        # create the real roller with a custom Indexer
        # from the attributes of the old roller.
        # After next line, all the attributes (public and private)
        #    of `_roller` are accessible on self
        self._roller = obj.rolling(
            window=self.window_indexer,
            min_periods=verified.min_periods,  # roller.min_periods
            win_type=verified._win_type,  # use private here to silence warning
            on=verified.on,
            center=verified.center,
            closed=closed,
            axis=verified.axis,
        )

    def _call_roll_func(self, name, args, kwargs):
        result = getattr(self._roller, name)(*args, **kwargs)
        if self._dtlike_window and not self._expand:
            result = self._remove_expanding_ramps(result)
        return result

    def _remove_expanding_ramps(self, result):
        if len(result.index) == 0:
            return result

        index = self._index_array
        window_size = self._win_freq_i8
        decreasing = index[0] > index[-1]
        mask = np.full_like(index, False, dtype=bool)

        if self.center:
            window_size //= 2

        if self.center or self._forward:
            if decreasing:
                mask |= index < index[-1] + window_size
            else:
                mask |= index > index[-1] - window_size

        if self.center or not self._forward:
            if decreasing:
                mask |= index > index[0] - window_size
            else:
                mask |= index < index[0] + window_size

        if window_size > 0:
            result[mask] = np.nan

        return result

    # =========================================================================
    # public interface
    #
    # All attributes of roller are accessible on self.
    # Public attributes are listed below, for the only reason to provide an
    # autocompletion for them.
    # Private attributes are `wrapped` via ``__getattr__``
    # =========================================================================

    def __getattr__(self, item):
        return getattr(self._roller, item)

    obj = AttrWrapper("obj")
    closed = AttrWrapper("closed")
    center = AttrWrapper("center")
    axis = AttrWrapper("axis")
    exclusions = AttrWrapper("exclusions")
    is_datetimelike = AttrWrapper("is_datetimelike")
    is_freq_type = AttrWrapper("is_freq_type")
    min_periods = AttrWrapper("min_periods")
    ndim = AttrWrapper("ndim")
    on = AttrWrapper("on")
    sparse = AttrWrapper("sparse")
    win_freq = AttrWrapper("win_freq")
    win_type = AttrWrapper("win_type")
    method = AttrWrapper("method")

    def sum(self, *args, **kwargs):
        return self._call_roll_func("sum", args, kwargs)

    def count(self, *args, **kwargs):
        return self._call_roll_func("count", args, kwargs)

    def mean(self, *args, **kwargs):
        return self._call_roll_func("mean", args, kwargs)

    def median(self, *args, **kwargs):
        return self._call_roll_func("median", args, kwargs)

    def min(self, *args, **kwargs):
        return self._call_roll_func("min", args, kwargs)

    def max(self, *args, **kwargs):
        return self._call_roll_func("max", args, kwargs)

    def skew(self, *args, **kwargs):
        return self._call_roll_func("skew", args, kwargs)

    def kurt(self, *args, **kwargs):
        return self._call_roll_func("kurt", args, kwargs)

    def var(self, *args, **kwargs):
        return self._call_roll_func("var", args, kwargs)

    def std(self, *args, **kwargs):
        return self._call_roll_func("std", args, kwargs)

    def sem(self, *args, **kwargs):
        return self._call_roll_func("sem", args, kwargs)

    def quantile(self, *args, **kwargs):
        return self._call_roll_func("quantile", args, kwargs)

    def cov(self, *args, **kwargs):
        return self._call_roll_func("cov", args, kwargs)

    def corr(self, *args, **kwargs):
        return self._call_roll_func("corr", args, kwargs)

    def apply(self, *args, **kwargs):
        return self._call_roll_func("apply", args, kwargs)

    def aggregate(self, *args, **kwargs):
        return self._call_roll_func("aggregate", args, kwargs)

    agg = aggregate

    def validate(self):  # dummy function to fit Rolling class
        self._roller.validate()


customRoller = CustomRoller



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Union, Callable

import numpy as np
import pandas as pd

from saqc.constants import BAD


class Rolling:
    def roll(
        self,
        field: str,
        window: Union[str, int],
        func: Callable[[pd.Series], np.ndarray] = np.mean,
        min_periods: int = 0,
        center: bool = True,
        **kwargs
    ):
        return self._defer("roll", locals())



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import pandas as pd
from dios.dios.dios import DictOfSeries

from saqc.constants import BAD, UNFLAGGED, FILTER_ALL
from saqc.core.flags import Flags
from saqc import SaQC
from saqc.lib.tools import toSequence

from tests.common import initData


@pytest.fixture
def data():
    return initData()


def test_writeTargetFlagGeneric(data):
    params = [
        (["tmp"], lambda x, y: pd.Series(True, index=x.index.union(y.index))),
        (
            ["tmp1", "tmp2"],
            lambda x, y: [pd.Series(True, index=x.index.union(y.index))] * 2,
        ),
    ]
    for targets, func in params:
        expected_meta = {
            "func": "flagGeneric",
            "args": (),
            "kwargs": {
                "field": data.columns.tolist(),
                "func": func.__name__,
                "target": targets,
                "flag": BAD,
                "dfilter": FILTER_ALL,
            },
        }

        saqc = SaQC(data=data)
        saqc = saqc.flagGeneric(field=data.columns, target=targets, func=func, flag=BAD)
        for target in targets:
            assert saqc._flags.history[target].hist.iloc[0].tolist() == [BAD]
            assert saqc._flags.history[target].hist.iloc[0].tolist() == [BAD]
            assert saqc._flags.history[target].meta[0] == expected_meta


def test_overwriteFieldFlagGeneric(data):
    params = [
        (["var1"], lambda x: pd.Series(True, index=x.index)),
        (
            ["var1", "var2"],
            lambda x, y: [pd.Series(True, index=x.index.union(y.index))] * 2,
        ),
    ]

    flag = 12

    for fields, func in params:
        expected_meta = {
            "func": "flagGeneric",
            "args": (),
            "kwargs": {
                "field": fields,
                "target": fields,
                "func": func.__name__,
                "flag": flag,
                "dfilter": FILTER_ALL,
            },
        }

        saqc = SaQC(
            data=data.copy(),
            flags=Flags(
                {
                    k: pd.Series(data[k] % 2, index=data[k].index).replace(
                        {0: UNFLAGGED, 1: 127}
                    )
                    for k in data.columns
                }
            ),
        )

        res = saqc.flagGeneric(field=fields, func=func, flag=flag)
        for field in fields:
            assert (data[field] == res.data[field]).all(axis=None)
            histcol0 = res._flags.history[field].hist[0]
            histcol1 = res._flags.history[field].hist[1]
            assert (histcol1[histcol0 == 127.0].isna()).all()
            assert (histcol1[histcol0 != 127.0] == flag).all()
            assert res._flags.history[field].meta[0] == {}
            assert res._flags.history[field].meta[1] == expected_meta


def test_writeTargetProcGeneric(data):
    fields = ["var1", "var2"]
    params = [
        (["tmp"], lambda x, y: x + y),
        (["tmp1", "tmp2"], lambda x, y: (x + y, y * 2)),
    ]
    dfilter = 128
    for targets, func in params:

        expected_data = DictOfSeries(
            func(*[data[f] for f in fields]), columns=toSequence(targets)
        ).squeeze()

        expected_meta = {
            "func": "procGeneric",
            "args": (),
            "kwargs": {
                "field": data.columns.tolist(),
                "target": targets,
                "func": func.__name__,
                "flag": BAD,
                "dfilter": dfilter,
            },
        }
        saqc = SaQC(
            data=data,
            flags=Flags(
                {k: pd.Series(127.0, index=data[k].index) for k in data.columns}
            ),
        )
        res = saqc.processGeneric(
            field=fields, target=targets, func=func, flag=BAD, dfilter=dfilter
        )
        assert (expected_data == res.data[targets].squeeze()).all(axis=None)
        # check that new histories where created
        for target in targets:
            assert res._flags.history[target].hist.iloc[0].tolist() == [BAD]
            assert res._flags.history[target].meta[0] == expected_meta


def test_overwriteFieldProcGeneric(data):
    params = [
        (["var1"], lambda x: x * 2),
        (["var1", "var2"], lambda x, y: (x + y, y * 2)),
    ]
    dfilter = 128
    flag = 12
    for fields, func in params:
        expected_data = DictOfSeries(
            func(*[data[f] for f in fields]), columns=fields
        ).squeeze()

        expected_meta = {
            "func": "procGeneric",
            "args": (),
            "kwargs": {
                "field": fields,
                "target": fields,
                "func": func.__name__,
                "flag": flag,
                "dfilter": dfilter,
            },
        }

        saqc = SaQC(
            data=data,
            flags=Flags(
                {k: pd.Series(127.0, index=data[k].index) for k in data.columns}
            ),
        )

        res = saqc.processGeneric(field=fields, func=func, flag=flag, dfilter=dfilter)
        assert (expected_data == res.data[fields].squeeze()).all(axis=None)
        # check that the histories got appended
        for field in fields:
            assert (res._flags.history[field].hist[0] == 127.0).all()
            assert (res._flags.history[field].hist[1] == 12.0).all()
            assert res._flags.history[field].meta[0] == {}
            assert res._flags.history[field].meta[1] == expected_meta



from . import pandas_bridge as pdextra
from .base import (
    _DiosBase,
    _is_dios_like,
    _is_bool_dios_like,
)

import numpy as np
import pandas as pd


class _Indexer:
    def __init__(self, obj: _DiosBase):
        self.obj = obj
        self._data = obj._data

    def _unpack_key(self, key):

        key = list(key) if pdextra.is_iterator(key) else key

        if isinstance(key, tuple):
            if len(key) > 2:
                raise KeyError("To many indexers")
            rowkey, colkey = key
        else:
            rowkey, colkey = key, slice(None)

        if isinstance(rowkey, tuple) or isinstance(colkey, tuple):
            raise KeyError(f"{key}. tuples are not allowed.")

        rowkey = list(rowkey) if pdextra.is_iterator(rowkey) else rowkey
        colkey = list(colkey) if pdextra.is_iterator(colkey) else colkey
        return rowkey, colkey

    def _set_value_muli_column(self, rowkey, colkey, value, xloc="loc"):
        """set value helper for loc and iloc"""

        data = getattr(self._data, xloc)[colkey]

        hashable_rkey = pdextra.is_hashable(rowkey)
        dioslike_value = False
        iter_value = False

        if _is_dios_like(value):
            dioslike_value = True
            if hashable_rkey:
                raise ValueError(f"Incompatible indexer with DictOfSeries")

        elif pdextra.is_list_like(value):
            value = value.values if isinstance(value, pd.Series) else value
            iter_value = True
            if len(value) != len(data):
                raise ValueError(
                    f"shape mismatch: value array of shape (.., {len(value)}) could "
                    f"not be broadcast to indexing result of shape (.., {len(data)})"
                )
        c = "?"
        try:
            for i, c in enumerate(data.index):
                dat = data.at[c]
                dat_xloc = getattr(dat, xloc)

                if dioslike_value:
                    # set to empty series fail; emptySer.loc[:] = [2,1]
                    # len(scalar) -> would fail, but cannot happen,
                    # because dioslike+hashable, already was checked
                    if len(dat_xloc[rowkey]) == 0:
                        continue

                # unpack the value if necessary
                if iter_value:
                    val = value[i]
                elif dioslike_value:
                    val = value[c] if c in value else np.nan
                else:
                    val = value

                dat_xloc[rowkey] = val

        except Exception as e:
            raise type(e)(f"failed for column {c}: " + str(e)) from e


# #############################################################################


class _LocIndexer(_Indexer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):

        rowkey, colkey = self._unpack_key(key)
        if _is_dios_like(rowkey) or _is_dios_like(colkey):
            raise ValueError("Could not index with multidimensional key")

        # simple optimisation
        if pdextra.is_null_slice(rowkey) and pdextra.is_null_slice(colkey):
            return self.obj.copy()

        data = self._data.loc[colkey].copy()

        # .loc[any, scalar] -> (a single) series
        # .loc[scalar, scalar] -> (a single) value
        if pdextra.is_hashable(colkey):
            new = data.loc[rowkey]

        # .loc[any, non-scalar]
        else:
            k = "?"
            try:

                for k in data.index:
                    data.at[k] = data.at[k].loc[rowkey]

            except Exception as e:
                raise type(e)(f"failed for column {k}: " + str(e)) from e

            # .loc[scalar, non-scalar] -> column-indexed series
            if pdextra.is_hashable(rowkey):
                new = data

            # .loc[non-scalar, non-scalar] -> dios
            else:
                new = self.obj.copy_empty(columns=False)
                new._data = data

        return new

    def __setitem__(self, key, value):

        rowkey, colkey = self._unpack_key(key)
        if _is_dios_like(rowkey) or _is_dios_like(colkey):
            raise ValueError("Cannot index with multi-dimensional key")

        # .loc[any, scalar] - set on single column
        if pdextra.is_hashable(colkey):

            # .loc[dont-care, new-scalar] = val
            if colkey not in self.obj.columns:
                self.obj._insert(colkey, value)

            # .loc[any, scalar] = multi-dim
            elif _is_dios_like(value) or pdextra.is_nested_list_like(value):
                raise ValueError("Incompatible indexer with multi-dimensional value")

            # .loc[any, scalar] = val
            else:
                self._data.at[colkey].loc[rowkey] = value

        # .loc[any, non-scalar] = any
        else:
            self._set_value_muli_column(rowkey, colkey, value, xloc="loc")


# #############################################################################


class _iLocIndexer(_Indexer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getitem__(self, key):
        rowkey, colkey = self._unpack_key(key)
        if _is_dios_like(rowkey) or _is_dios_like(colkey):
            raise ValueError("Cannot index with multidimensional key")

        # simple optimisation
        if pdextra.is_null_slice(rowkey) and pdextra.is_null_slice(colkey):
            return self.obj.copy()

        data = self._data.iloc[colkey].copy()

        # .iloc[any, int] -> single series
        # .iloc[int, int] -> single value
        if pdextra.is_integer(colkey):
            new = data.iloc[rowkey]

        # .iloc[any, non-int]
        else:
            k = "?"
            try:

                for k in data.index:
                    data.at[k] = data.at[k].iloc[rowkey]

            except Exception as e:
                raise type(e)(f"failed for column {k}: " + str(e)) from e

            # .iloc[int, non-int] -> column-indexed series
            if pdextra.is_integer(rowkey):
                new = data

            # .iloc[non-int, non-int] -> dios
            else:
                new = self.obj.copy_empty(columns=False)
                new._data = data

        return new

    def __setitem__(self, key, value):
        rowkey, colkey = self._unpack_key(key)
        if _is_dios_like(rowkey) or _is_dios_like(colkey):
            raise ValueError("Cannot index with multidimensional key")

        # .iloc[any, int] = Any
        if pdextra.is_integer(colkey):
            if _is_dios_like(value) or pdextra.is_nested_list_like(value):
                raise ValueError("Incompatible indexer with multi-dimensional value")
            self._data.iat[colkey].iloc[rowkey] = value

        # .iloc[any, non-int] = Any
        else:
            self._set_value_muli_column(rowkey, colkey, value, xloc="iloc")


# #############################################################################


class _aLocIndexer(_Indexer):
    """align Indexer

    Automatically align (alignable) indexer on all possible axis,
    and handle indexing with non-existent or missing keys gracefully.

    Also align (alignable) values before setting them with .loc
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._usebool = True

    def __call__(self, usebool=True):
        """We are called if the user want to set `usebool=False', which make
        boolean alignable indexer treat as non-boolean alignable indexer.

        Explanation: A boolean dios indexer align its indices with the indices
        of the receiving dios like a non-boolean dios indexer also would do.
        Additionally all rows with False values are kicked too. To disable
         that `usebool=False` can be given."""
        self._usebool = usebool
        return self

    def __getitem__(self, key):
        rowkeys, colkeys, lowdim = self._unpack_key_aloc(key)
        data = pd.Series(dtype="O", index=colkeys)
        kws = dict(itype=self.obj.itype, cast_policy=self.obj._policy)

        c = "?"
        try:

            for i, c in enumerate(data.index):
                data.at[c] = self._data.at[c].loc[rowkeys[i]]

        except Exception as e:
            raise type(e)(f"failed for column {c}: " + str(e)) from e

        if lowdim:
            return data.squeeze()
        else:
            return self.obj._constructor(data=data, fastpath=True, **kws)._finalize(
                self.obj
            )

    def __setitem__(self, key, value):
        rowkeys, colkeys, _ = self._unpack_key_aloc(key)

        def iter_self(colkeys, position=False):
            c = "?"
            try:

                for i, c in enumerate(colkeys):
                    dat = self._data.at[c]
                    rk = rowkeys[i]
                    if len(dat.loc[rk]) == 0:
                        continue
                    yield dat, rk, i if position else c

            except Exception as e:
                raise type(e)(f"failed for column {c}: " + str(e)) from e

        # align columns, for rows use series.loc to align
        if _is_dios_like(value):
            colkeys = value.columns.intersection(colkeys)
            for dat, rk, c in iter_self(colkeys):
                dat.loc[rk] = value[c]

        # no align, no merci
        elif pdextra.is_nested_list_like(value):
            if len(colkeys) != len(value):
                raise ValueError(
                    f"shape mismatch: values array of shape "
                    f"(.., {len(value)}) could not "
                    f"be broadcast to indexing result of "
                    f"shape (.., {len(colkeys)})"
                )
            for dat, rk, i in iter_self(colkeys, position=True):
                dat.loc[rk] = value[i]

        # align rows by using series.loc
        elif isinstance(value, pd.Series):
            for dat, rk, _ in iter_self(colkeys):
                dat.loc[rk] = value

        # no align, no merci
        else:
            for dat, rk, _ in iter_self(colkeys):
                dat.loc[rk] = value

    def _unpack_key_aloc(self, key):
        """
        Return a list of row indexer and a list of existing(!) column labels.
        Both list always have the same length and also could be empty together.

        Note:
            The items of the row indexer list should be passed to pd.Series.loc[]
        """
        # if a single column-key is given, the caller may
        # want to return a single Series, instead of a dios
        lowdim = False

        def keys_from_bool_dios_like(key):
            if not _is_bool_dios_like(key):
                raise ValueError("Must pass dios-like key with boolean values only.")
            colkey = self.obj.columns.intersection(key.columns)
            rowkey = []
            for c in colkey:
                b = key[c]
                rowkey += [self._data.at[c].index.intersection(b[b].index)]
            return rowkey, colkey, lowdim

        def keys_from_dios_like(key):
            colkey = self.obj.columns.intersection(key.columns)
            rowkey = [self._data.at[c].index.intersection(key[c].index) for c in colkey]
            return rowkey, colkey, lowdim

        rowkey, colkey = self._unpack_key(key)

        if _is_dios_like(colkey) or pdextra.is_nested_list_like(colkey):
            raise ValueError("Could not index with multi-dimensional column key.")

        # giving the ellipsis as column key, is an alias
        # for giving `usebool=False`. see self.__call__()
        if colkey is Ellipsis:
            self._usebool = False
            colkey = slice(None)

        # .aloc[dios]
        if _is_dios_like(rowkey):

            if not pdextra.is_null_slice(colkey):
                raise ValueError(
                    f"Could not index with a dios-like indexer as rowkey,"
                    f"and a column key of that type {type(colkey)}"
                )
            if self._usebool:
                return keys_from_bool_dios_like(rowkey)
            else:
                return keys_from_dios_like(rowkey)

        # handle gracefully: scalar
        elif pdextra.is_hashable(colkey):
            colkey = [colkey] if colkey in self.obj.columns else []
            lowdim = True

        # column-alignable: list-like, filter only existing columns
        elif pdextra.is_list_like(colkey) and not pdextra.is_bool_indexer(colkey):
            colkey = colkey.values if isinstance(colkey, pd.Series) else colkey
            colkey = self.obj.columns.intersection(colkey)

        # handle gracefully (automatically)
        # just a simple optimisation
        elif pdextra.is_null_slice(colkey):
            colkey = self.obj.columns

        # not alignable, fall back to .loc (boolean list/series, slice(..), etc.
        else:
            colkey = self._data.loc[colkey].index

        if len(colkey) == 0:  # (!) `if not colkey:` fails for pd.Index
            return [], [], lowdim

        rowkey = self._get_rowkey(rowkey, colkey)

        return rowkey, colkey, lowdim

    def _get_rowkey(self, rowkey, colkey, depth=0):

        if pdextra.is_nested_list_like(rowkey) and depth == 0:
            rowkey = rowkey.values if isinstance(rowkey, pd.Series) else rowkey
            if len(rowkey) != len(colkey):
                raise ValueError(
                    "Nested arrays indexer must have same (outer) "
                    "length than the number of selected columns."
                )
            indexer = []
            for i, c in enumerate(colkey):
                # recurse to get the row indexer from inner element
                indexer += self._get_rowkey(rowkey[i], [c], depth=depth + 1)
            rowkey = indexer

        # row-alignable: pd.Series(), align rows to every series in colkey (columns)
        elif isinstance(rowkey, pd.Series):
            if self._usebool and pdextra.is_bool_indexer(rowkey):
                rowkey = [
                    self._data.at[c].index.intersection(rowkey[rowkey].index)
                    for c in colkey
                ]
            else:
                rowkey = [
                    self._data.at[c].index.intersection(rowkey.index) for c in colkey
                ]

        # handle gracefully: scalar, transform to row-slice
        elif pdextra.is_hashable(rowkey):
            rowkey = [slice(rowkey, rowkey)] * len(colkey)

        # handle gracefully: list-like, filter only existing rows
        # NOTE: dios.aloc[series.index] is processed here
        elif pdextra.is_list_like(rowkey) and not pdextra.is_bool_indexer(rowkey):
            rowkey = [self._data.at[c].index.intersection(rowkey) for c in colkey]

        # not alignable
        # the rowkey is processed by .loc someway in
        # the calling function - (eg. slice(..), boolean list-like, etc.)
        else:
            rowkey = [rowkey] * len(colkey)

        return rowkey


# #############################################################################


class _AtIndexer(_Indexer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_key(self, key):
        if not (
            isinstance(key, tuple)
            and len(key) == 2
            and pdextra.is_hashable(key[0])
            and pdextra.is_hashable(key[1])
        ):
            raise KeyError(
                f"{key}. `.at` takes exactly one scalar row-key "
                "and one scalar column-key"
            )

    def __getitem__(self, key):
        self._check_key(key)
        return self._data.at[key[1]].at[key[0]]

    def __setitem__(self, key, value):
        self._check_key(key)
        if _is_dios_like(value) or pdextra.is_nested_list_like(value):
            raise TypeError(
                ".at[] cannot be used to set multi-dimensional values, use .aloc[] instead."
            )
        self._data.at[key[1]].at[key[0]] = value


# #############################################################################


class _iAtIndexer(_Indexer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _check_key(self, key):
        if not (
            isinstance(key, tuple)
            and len(key) == 2
            and pdextra.is_integer(key[0])
            and pdextra.is_integer(key[1])
        ):
            raise KeyError(
                f"{key} `.iat` takes exactly one integer positional "
                f"row-key and one integer positional scalar column-key"
            )

    def __getitem__(self, key):
        self._check_key(key)
        return self._data.iat[key[1]].iat[key[0]]

    def __setitem__(self, key, value):
        self._check_key(key)
        if _is_dios_like(value) or pdextra.is_nested_list_like(value):
            raise TypeError(
                ".iat[] cannot be used to set multi-dimensional values, use .aloc[] instead."
            )
        self._data.iat[key[1]].iat[key[0]] = value



#!/usr/bin/env python
import pytest

from .test_setup import *


__author__ = "Bert Palm"
__email__ = "bert.palm@ufz.de"
__copyright__ = "Copyright 2018, Helmholtz-Centrum für Umweltforschung GmbH - UFC"


@pytest.mark.parametrize("left", diosFromMatr(DATA_ALIGNED))
@pytest.mark.parametrize("right", diosFromMatr(DATA_ALIGNED))
def test__eq__(left, right):
    a, b = left, right
    _test = a == b
    for c in _test:
        for i in _test[c].index:
            res = (_test[c])[i]
            e1 = a[c][i]
            e2 = b[c][i]
            exp = e1 == e2
            assert res == exp


@pytest.mark.filterwarnings("ignore: invalid value encountered in long_scalars")
@pytest.mark.filterwarnings("ignore: divide by zero encountered in long_scalars")
@pytest.mark.parametrize("left", diosFromMatr(DATA_ALIGNED))
@pytest.mark.parametrize("right", diosFromMatr(DATA_ALIGNED))
@pytest.mark.parametrize("op", OP2)
def test__op2__aligningops(left, right, op):
    a, b = left, right
    test = op(a, b)
    for c in test:
        for j in test[c].index:
            exp = op(a[c][j], b[c][j])
            res = test[c][j]
            if not np.isfinite(res):
                print(f"\n\n{a[c][j]} {OP_MAP[op]} {b[c][j]}")
                print(f"\nres: {res}, exp:{exp}, op: {OP_MAP[op]}")
                pytest.skip("test not support non-finite values")
                return
            assert res == exp


@pytest.mark.filterwarnings("ignore: invalid value encountered in long_scalars")
@pytest.mark.filterwarnings("ignore: divide by zero encountered in long_scalars")
@pytest.mark.parametrize("left", diosFromMatr(DATA_UNALIGNED))
@pytest.mark.parametrize("right", diosFromMatr(DATA_UNALIGNED))
@pytest.mark.parametrize("op", OPNOCOMP)
def test__op2__UNaligningops(left, right, op):
    try:
        a, b = left, right
        test = op(a, b)
        for c in test:
            for j in test[c].index:
                exp = op(a[c][j], b[c][j])
                res = test[c][j]
                if not np.isfinite(res):
                    print(f"\n\n{a[c][j]} {OP_MAP[op]} {b[c][j]}")
                    print(f"\nres: {res}, exp:{exp}, op: {OP_MAP[op]}")
                    pytest.skip("test not support non-finite values")
                    return
                assert res == exp
    except ZeroDivisionError:
        pytest.skip("ZeroDivisionError")


@pytest.mark.parametrize("data", diosFromMatr(ALL))
@pytest.mark.parametrize("op", OP1)
def test__op1__(data, op):
    test = op(data)
    res = [entry for col in test for entry in test[col]]
    e = [entry for col in data for entry in data[col]]
    for i in range(len(res)):
        exp = op(e[i])
        assert res[i] == exp



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Optional
from typing_extensions import Literal

import saqc
import numpy as np

from saqc.constants import FILTER_NONE


class Tools:
    def copyField(self, field: str, target: str, **kwargs) -> saqc.SaQC:
        return self._defer("copyField", locals())

    def dropField(self, field: str, **kwargs) -> saqc.SaQC:
        return self._defer("dropField", locals())

    def renameField(self, field: str, new_name: str, **kwargs) -> saqc.SaQC:
        return self._defer("renameField", locals())

    def maskTime(
        self,
        field: str,
        mode: Literal["periodic", "mask_field"],
        mask_field: Optional[str] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        closed: bool = True,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("maskTime", locals())

    def plot(
        self,
        field: str,
        path: Optional[str] = None,
        max_gap: Optional[str] = None,
        stats: bool = False,
        history: Optional[Literal["valid", "complete", "clear"]] = "valid",
        xscope: Optional[slice] = None,
        phaseplot: Optional[str] = None,
        stats_dict: Optional[dict] = None,
        store_kwargs: Optional[dict] = None,
        dfilter: Optional[float] = FILTER_NONE,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("plot", locals())



#!/usr/bin/env python
import pytest
from .test_setup import *

import pandas as pd
from pandas.core.dtypes.common import is_dict_like, is_nested_list_like
import numpy as np
from copy import deepcopy

__author__ = "Bert Palm"
__email__ = "bert.palm@ufz.de"
__copyright__ = "Copyright 2018, Helmholtz-Zentrum für Umweltforschung GmbH - UFZ"


arr = np.random.rand(8)
TESTDATA = [
    None,  # empty  # 0
    [1],  # 1
    arr.copy(),  # 2
    np.array([arr.copy(), arr.copy(), arr.copy()]),  # 3 - nested list
    range(4),  # 4
    dict(a=arr.copy(), b=arr.copy()),  # 5 dict
    pd.DataFrame(dict(a=arr.copy(), b=arr.copy())),  # 6 df
]


@pytest.mark.parametrize("data", TESTDATA)
@pytest.mark.parametrize("with_column_param", [False, True])
def test_dios_create(data, with_column_param):

    data_copy0 = deepcopy(data)
    data_copy1 = deepcopy(data)

    # create columns list
    if with_column_param:
        df = pd.DataFrame(data=data_copy0)
        col = [f"new_{c}" for c in df]
    else:
        col = None

    if is_nested_list_like(data):
        # giving nested lists, work different between df and dios
        data_copy1 = data_copy1.transpose()

    df = pd.DataFrame(data=data_copy0, columns=col)
    dios = DictOfSeries(data=data_copy1, columns=col)

    assert dios.columns.equals(df.columns)

    eq, msg = dios_eq_df(dios, df, with_msg=True)
    assert eq, msg



from builtins import range

from dios import *
import numpy as np

if __name__ == "__main__":
    dios_options[OptsFields.mixed_itype_warn_policy] = Opts.itype_warn
    print(dios_options)

    df = pd.DataFrame(columns=range(1000))
    pd.Series()
    # print(df)
    # exit(99)

    # dios_options[OptsFields.disp_max_cols] = 5
    # dios_options[OptsFields.disp_max_rows] = 100
    dios_options[OptsFields.disp_min_rows] = 50
    # dios_options[OptsFields.dios_repr] = Opts.repr_aligned

    n = 10
    d = DictOfSeries(
        dict(
            l=pd.Series(0, index=range(0, 30)),
            # i123=pd.Series(dtype='O'),
            a=pd.Series(1, index=range(0, n)),
            nan=pd.Series(np.nan, index=range(3, n + 3)),
            b=pd.Series(2, index=range(0, n * 2, 2)),
            c=pd.Series(3, index=range(n, n * 2)),
            d=pd.Series(4, index=range(-n // 2, n // 2)),
            # z=pd.Series([1, 2, 3], index=list("abc"))
        )
    )

    def f(s):
        sec = 10 ** 9
        s.index = pd.to_datetime(s.index * sec)
        return s

    dd = d.apply(f)
    print(d)

    # print(d.to_df())
    # print(pd.options.display.max_rows)
    # print(d.to_str(col_delim=' | ', col_space=20, header_delim='0123456789'))
    # print(d.to_str(col_delim=' | ', col_space=20, max_cols=4 ))
    di = DictOfSeries(columns=[])
    print(di)
    # print(DictOfSeries(data=1, columns=['a']))



#!/usr/bin/env python
"""
The module comprises flag value constants in use throughout saqc.
The constants order as follows (from "worse" to "best"):

:py:const:`~saqc.constants.BAD` > :py:const:`~saqc.constants.DOUBTFUL` > :py:const:`~saqc.constants.GOOD` >
:py:const:`~saqc.constants.UNFLAGGED`
"""

__all__ = [
    "UNFLAGGED",
    "DOUBTFUL",
    "BAD",
    "GOOD",
    "ENVIRONMENT",
    "FILTER_ALL",
    "FILTER_NONE",
]


import numpy as np
import scipy.stats as st
import saqc.lib.ts_operators as ts_ops

# ----------------------------------------------------------------------
# global flag constants
# ----------------------------------------------------------------------

#: A :py:mod:`flag level constant <saqc.constants>`
#: , evaluating to the level, that indicates, no flag has been assigned to yet.
UNFLAGGED = -np.inf

#: A :py:mod:`flag level constant <saqc.constants>`
#: , evaluating to the lowest level level of flagging, that is
#: not :py:const:`UNFLAGGED <saqc.constants.UNFLAGGED>`.
GOOD = 0

#: A :py:mod:`flag level constant <saqc.constants>`
#: , evaluating to a somewhat modest flag level.
DOUBTFUL = 25.0

#: A :py:mod:`flag level constant <saqc.constants>`
#: , evaluating to the highest (internal) flag level available.
BAD = 255.0


# ----------------------------------------------------------------------
# global dfilter constants
# ----------------------------------------------------------------------

#: A :py:mod:`dfilter constant <saqc.constants>`
#: , mask/filter all flagged data.
FILTER_ALL = -np.inf

#: A :py:mod:`dfilter constant <saqc.constants>`
#: , mask/filter no data at all.
FILTER_NONE = np.inf


# ----------------------------------------------------------------------
# other
# ----------------------------------------------------------------------

#: A :py:mod:`flag level constant <saqc.constants>`
ENVIRONMENT = {
    # Not A number Constant.
    "NAN": np.nan,
    # Pointwise absolute Value Function.
    "abs": np.abs,
    # Maximum Value Function. Ignores NaN.
    "max": np.nanmax,
    # Minimum Value Function. Ignores NaN.
    "min": np.nanmin,
    # Mean Value Function. Ignores NaN.
    "mean": np.nanmean,
    # Summation. Ignores NaN.
    "sum": np.nansum,
    # Standart Deviation. Ignores NaN.
    "len": len,
    # Pointwise Exponential.
    "exp": np.exp,
    # Pointwise Logarithm.
    "log": np.log,
    # Logarithm, returning NaN for zero input, instead of -inf.
    "nanLog": ts_ops.zeroLog,
    # Standart Deviation. Ignores NaN.
    "std": np.nanstd,
    # Variance. Ignores NaN.
    "var": np.nanvar,
    # Median. Ignores NaN.
    "median": np.nanmedian,
    # Count Number of values. Ignores NaNs.
    "count": ts_ops.count,
    # Identity.
    "id": ts_ops.identity,
    # Returns a Series` diff.
    "diff": ts_ops.difference,
    # Scales data to [0,1] Interval.
    "scale": ts_ops.normScale,
    # Standardize with Standart Deviation.
    "zScore": lambda x: st.zscore(x, nan_policy="omit"),
    # Standardize with Median and MAD.
    "madScore": ts_ops.standardizeByMedian,
    # Standardize with Median and inter quantile range.
    "iqsScore": ts_ops.standardizeByIQR,
    "GOOD": GOOD,
    "BAD": BAD,
    "UNFLAGGED": UNFLAGGED,
    "DOUBTFUL": DOUBTFUL,
}



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Tuple, Union, Callable
from typing_extensions import Literal
import numpy as np
import pandas as pd
from dios import DictOfSeries

from saqc.constants import *
from saqc.core import register, Flags
from saqc.core.register import _isflagged, processing
from saqc.lib.ts_operators import interpolateNANs

_SUPPORTED_METHODS = Literal[
    "linear",
    "time",
    "nearest",
    "zero",
    "slinear",
    "quadratic",
    "cubic",
    "spline",
    "barycentric",
    "polynomial",
    "krogh",
    "piecewise_polynomial",
    "spline",
    "pchip",
    "akima",
]


@register(
    mask=["field"],
    demask=["field"],
    squeeze=[],  # func handles history by itself
)
def interpolateByRolling(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    window: Union[str, int],
    func: Callable[[pd.Series], float] = np.median,
    center: bool = True,
    min_periods: int = 0,
    flag: float = UNFLAGGED,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Interpolates nan-values in the data by assigning them the aggregation result of the window surrounding them.

    Parameters
    ----------
    data : dios.DictOfSeries
        The data container.

    field : str
        Name of the column, holding the data-to-be-interpolated.

    flags : saqc.Flags
        A flags object, holding flags and additional Information related to `data`.

    window : int, str
        The size of the window, the aggregation is computed from. An integer define the number of periods to be used,
        an string is interpreted as an offset. ( see `pandas.rolling` for more information).
        Integer windows may result in screwed aggregations if called on none-harmonized or irregular data.

    func : Callable
        The function used for aggregation.

    center : bool, default True
        Center the window around the value. Can only be used with integer windows, otherwise it is silently ignored.

    min_periods : int
        Minimum number of valid (not np.nan) values that have to be available in a window for its aggregation to be
        computed.

    flag : float or None, default UNFLAGGED
        Flag that is to be inserted for the interpolated values.
        If `None` the old flags are kept, even if the data is valid now.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
    """
    datcol = data[field]
    roller = datcol.rolling(window=window, center=center, min_periods=min_periods)
    try:
        func_name = func.__name__
        if func_name[:3] == "nan":
            func_name = func_name[3:]
        rolled = getattr(roller, func_name)()
    except AttributeError:
        rolled = roller.apply(func)

    na_mask = datcol.isna()
    interpolated = na_mask & rolled.notna()
    datcol[na_mask] = rolled[na_mask]
    data[field] = datcol

    new_col = pd.Series(np.nan, index=flags[field].index)
    new_col.loc[interpolated] = np.nan if flag is None else flag

    # todo kwargs must have all passed args except data,field,flags
    flags.history[field].append(
        new_col, {"func": "interpolateByRolling", "args": (), "kwargs": kwargs}
    )

    return data, flags


@register(
    mask=["field"],
    demask=["field"],
    squeeze=[],  # func handles history by itself
)
def interpolateInvalid(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    method: _SUPPORTED_METHODS,
    order: int = 2,
    limit: int = 2,
    downgrade: bool = False,
    flag: float = UNFLAGGED,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Function to interpolate nan values in the data.

    There are available all the interpolation methods from the pandas.interpolate method and they are applicable by
    the very same key words, that you would pass to the ``pd.Series.interpolate``'s method parameter.

    Parameters
    ----------
    data : dios.DictOfSeries
        The data container.

    field : str
        Name of the column, holding the data-to-be-interpolated.

    flags : saqc.Flags
        A flags object, holding flags and additional Information related to `data`.

    method : {"linear", "time", "nearest", "zero", "slinear", "quadratic", "cubic", "spline", "barycentric",
        "polynomial", "krogh", "piecewise_polynomial", "spline", "pchip", "akima"}
        The interpolation method to use.

    order : int, default 2
        If there your selected interpolation method can be performed at different 'orders' - here you pass the desired
        order.

    limit : int, default 2
        Maximum number of consecutive 'nan' values allowed for a gap to be interpolated. This really restricts the
        interpolation to chunks, containing not more than `limit` successive nan entries.

    flag : float or None, default UNFLAGGED
        Flag that is set for interpolated values. If ``None``, no flags are set at all.

    downgrade : bool, default False
        If `True` and the interpolation can not be performed at current order, retry with a lower order.
        This can happen, because the chosen ``method`` does not support the passed ``order``, or
        simply because not enough values are present in a interval.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
    """
    inter_data = interpolateNANs(
        data[field],
        method,
        order=order,
        inter_limit=limit,
        downgrade_interpolation=downgrade,
    )
    data[field] = inter_data

    interpolated = data[field].isna() & inter_data.notna()
    new_col = pd.Series(np.nan, index=flags[field].index)
    new_col.loc[interpolated] = np.nan if flag is None else flag

    # todo kwargs must have all passed args except data,field,flags
    flags.history[field].append(
        new_col, {"func": "interpolateInvalid", "args": (), "kwargs": kwargs}
    )

    return data, flags


def _resampleOverlapping(data: pd.Series, freq: str, fill_value):
    """TODO: docstring needed"""
    dtype = data.dtype
    end = data.index[-1].ceil(freq)
    data = data.resample(freq).max()
    data = data.combine(data.shift(1, fill_value=fill_value), max)
    if end not in data:
        data.loc[end] = fill_value
    return data.fillna(fill_value).astype(dtype)


@processing()
def interpolateIndex(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    freq: str,
    method: _SUPPORTED_METHODS,
    order: int = 2,
    limit: int = 2,
    downgrade: bool = False,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Function to interpolate the data at regular (equidistant) timestamps (or Grid points).

    Note, that the interpolation will only be calculated, for grid timestamps that have a preceding AND a succeeding
    valid data value within "freq" range.

    Parameters
    ----------
    data : dios.DictOfSeries
        The data container.

    field : str
        Name of the column, holding the data-to-be-interpolated.

    flags : saqc.Flags
        A flags object, holding flags and additional Information related to `data`.

    freq : str
        An Offset String, interpreted as the frequency of
        the grid you want to interpolate your data at.

    method : {"linear", "time", "nearest", "zero", "slinear", "quadratic", "cubic", "spline", "barycentric",
        "polynomial", "krogh", "piecewise_polynomial", "spline", "pchip", "akima"}: string
        The interpolation method you want to apply.

    order : int, default 2
        If there your selected interpolation method can be performed at different 'orders' - here you pass the desired
        order.

    limit : int, default 2
        Maximum number of consecutive 'nan' values allowed for a gap to be interpolated. This really restricts the
        interpolation to chunks, containing not more than `limit` successive nan entries.

    downgrade : bool, default False
        If `True` and the interpolation can not be performed at current order, retry with a lower order.
        This can happen, because the chosen ``method`` does not support the passed ``order``, or
        simply because not enough values are present in a interval.


    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values and shape may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
        Flags values and shape may have changed relatively to the flags input.
    """
    if data[field].empty:
        return data, flags

    datcol = data[field].copy()

    start, end = datcol.index[0].floor(freq), datcol.index[-1].ceil(freq)
    grid_index = pd.date_range(start=start, end=end, freq=freq, name=datcol.index.name)

    # todo:
    #  in future we could use `register(mask=[field], [], [])`
    #  and dont handle masking manually here
    flagged = _isflagged(flags[field], kwargs["dfilter"])

    # drop all points that hold no relevant grid information
    datcol = datcol[~flagged].dropna()

    # account for annoying case of subsequent frequency aligned values,
    # that differ exactly by the margin of 2*freq
    gaps = datcol.index[1:] - datcol.index[:-1] == 2 * pd.Timedelta(freq)
    gaps = datcol.index[1:][gaps]
    gaps = gaps.intersection(grid_index).shift(-1, freq)

    # prepare grid interpolation:
    datcol = datcol.reindex(datcol.index.union(grid_index))

    # do the grid interpolation
    inter_data = interpolateNANs(
        data=datcol,
        method=method,
        order=order,
        inter_limit=limit,
        downgrade_interpolation=downgrade,
    )

    # override falsely interpolated values:
    inter_data[gaps] = np.nan

    # store interpolated grid
    data[field] = inter_data[grid_index]

    history = flags.history[field].apply(
        index=data[field].index,
        func=_resampleOverlapping,
        func_kws=dict(freq=freq, fill_value=UNFLAGGED),
    )

    flags.history[field] = history
    return data, flags



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import numpy as np
import pandas as pd

import saqc
from saqc.core import initFlagsLike, SaQC, register
from saqc.core.flags import Flags
from saqc.core.register import processing

from tests.common import initData, flagAll

OPTIONAL = [False, True]


register(mask=["field"], demask=["field"], squeeze=["field"])(flagAll)


@pytest.fixture
def data():
    return initData(3)


@pytest.fixture
def flags(data, optional):
    if not optional:
        return initFlagsLike(data[data.columns[::2]]).toDios()


def test_errorHandling(data):
    @processing()
    def raisingFunc(data, field, flags, **kwargs):
        raise TypeError

    var1 = data.columns[0]

    with pytest.raises(TypeError):
        SaQC(data).raisingFunc(var1)


def test_duplicatedVariable():
    data = initData(1)
    var1 = data.columns[0]

    pflags = SaQC(data).flagDummy(var1).result.flags

    if isinstance(pflags.columns, pd.MultiIndex):
        cols = pflags.columns.get_level_values(0).drop_duplicates()
        assert np.all(cols == [var1])
    else:
        assert (pflags.columns == [var1]).all()


@pytest.mark.parametrize("optional", OPTIONAL)
def test_dtypes(data, flags):
    """
    Test if the categorical dtype is preserved through the core functionality
    """
    flags = initFlagsLike(data)
    flags_raw = flags.toDios()
    var1, var2 = data.columns[:2]

    pflags = SaQC(data, flags=flags_raw).flagAll(var1).flagAll(var2).result.flags_raw

    for c in pflags.columns:
        assert pflags[c].dtype == flags[c].dtype


def test_copy(data):
    qc = saqc.SaQC(data)

    qc = qc.flagRange("var1").flagRange("var1", min=0, max=0)

    deep = qc.copy(deep=True)
    shallow = qc.copy(deep=False)

    for copy in [deep, shallow]:
        assert copy is not qc
        assert copy._called is not qc._called
        assert copy._scheme is not qc._scheme
        assert copy._attrs is not qc._attrs

        assert copy._data is not qc._data
        assert copy._flags is not qc._flags

        assert copy._data._data is not qc._data._data
        assert copy._flags._data is not qc._flags._data

    # underling data copied
    assert deep._data._data.iloc[0] is not qc._data._data.iloc[0]
    assert (
        deep._flags._data["var1"].hist.index is not qc._flags._data["var1"].hist.index
    )

    # underling data NOT copied
    assert shallow._data._data.iloc[0] is qc._data._data.iloc[0]
    assert shallow._flags._data["var1"].hist.index is qc._flags._data["var1"].hist.index


def test_sourceTargetCopy():
    """
    test implicit copies
    """
    data = initData(1)
    var1 = data.columns[0]
    target = "new"

    @register(mask=["field"], demask=["field"], squeeze=["field"], handles_target=False)
    def flagTarget(data, field, flags, **kwargs):
        assert "target" not in kwargs
        return data, flags

    qc = SaQC(data, flags=Flags({var1: pd.Series(127.0, index=data[var1].index)}))
    qc = qc.flagTarget(field=var1, target=target)

    assert (qc.data[var1] == qc.data[target]).all(axis=None)
    assert all(qc.flags[var1] == qc.flags[target])


def test_sourceTargetNoCopy():
    """
    test bypassing of the imlpicit copy machiners
    """
    data = initData(1)
    var1 = data.columns[0]
    target = "new"

    @register(mask=["field"], demask=["field"], squeeze=["field"], handles_target=True)
    def flagField(data, field, flags, **kwargs):
        assert "target" in kwargs
        assert "target" not in data
        assert "target" not in flags
        return data, flags

    SaQC(data).flagField(field=var1, target=target)


def test_sourceTargetMultivariate():
    """
    test bypassing of the imlpicit copy machiners
    """
    data = initData(3)

    @register(
        mask=["field"],
        demask=["field"],
        squeeze=["field"],
        handles_target=True,
        multivariate=True,
    )
    def flagMulti(data, field, flags, **kwargs):
        assert "target" in kwargs
        assert "target" not in data
        assert "target" not in flags
        assert field == kwargs["target"]
        return data, flags

    SaQC(data).flagMulti(field=data.columns, target=data.columns)



#!/usr/bin/env python


import numbers
import dios
import numpy as np
import pandas as pd
from typing import get_type_hints
from contextlib import contextmanager

from hypothesis.strategies import (
    lists,
    sampled_from,
    composite,
    from_regex,
    sampled_from,
    datetimes,
    integers,
    register_type_strategy,
    from_type,
)
from hypothesis.extra.numpy import arrays, from_dtype
from hypothesis.strategies._internal.types import _global_type_lookup

from saqc.constants import *
from saqc.core.register import FUNC_MAP
from saqc.core import initFlagsLike

MAX_EXAMPLES = 50
# MAX_EXAMPLES = 100000


@composite
def dioses(draw, min_cols=1):
    """
    initialize data according to the current restrictions
    """
    # NOTE:
    # The following restriction showed up and should be enforced during init:
    # - Column names need to satisify the following regex: [A-Za-z0-9_-]+
    # - DatetimeIndex needs to be sorted
    # - Integer values larger than 2**53 lead to numerical instabilities during
    #   the integer->float->integer type conversion in _maskData/_unmaskData.

    cols = draw(lists(columnNames(), unique=True, min_size=min_cols))
    columns = {c: draw(dataSeries(min_size=3)) for c in cols}
    return dios.DictOfSeries(columns)


@composite
def dataSeries(
    draw, min_size=0, max_size=100, dtypes=("float32", "float64", "int32", "int64")
):
    if np.isscalar(dtypes):
        dtypes = (dtypes,)

    dtype = np.dtype(draw(sampled_from(dtypes)))
    if issubclass(dtype.type, numbers.Integral):
        info = np.iinfo(dtype)
    elif issubclass(dtype.type, numbers.Real):
        info = np.finfo(dtype)
    else:
        raise ValueError("only numerical dtypes are supported")
    # we don't want to fail just because of overflows
    elements = from_dtype(dtype, min_value=info.min + 1, max_value=info.max - 1)

    index = draw(daterangeIndexes(min_size=min_size, max_size=max_size))
    values = draw(arrays(dtype=dtype, elements=elements, shape=len(index)))
    return pd.Series(data=values, index=index)


@composite
def columnNames(draw):
    return draw(from_regex(r"[A-Za-z0-9_-]+", fullmatch=True))


@composite
def flagses(draw, data):
    """
    initialize a flags and set some flags
    """
    flags = initFlagsLike(data)
    for col, srs in data.items():
        loc_st = lists(
            sampled_from(sorted(srs.index)), unique=True, max_size=len(srs) - 1
        )
        flags[draw(loc_st), col] = BAD
    return flags


@composite
def functions(draw, module: str = None):
    samples = tuple(FUNC_MAP.values())
    if module:
        samples = tuple(f for f in samples if f.name.startswith(module))
    # samples = [FUNC_MAP["drift.correctExponentialDrift"]]
    return draw(sampled_from(samples))


@composite
def daterangeIndexes(draw, min_size=0, max_size=100):
    min_date = pd.Timestamp("1900-01-01").to_pydatetime()
    max_date = pd.Timestamp("2099-12-31").to_pydatetime()
    start = draw(datetimes(min_value=min_date, max_value=max_date))
    periods = draw(integers(min_value=min_size, max_value=max_size))
    freq = draw(sampled_from(["D", "H", "T", "min", "S", "L", "ms", "U", "us", "N"]))
    return pd.date_range(start, periods=periods, freq=freq)


@composite
def frequencyStrings(draw, _):
    freq = draw(sampled_from(["D", "H", "T", "min", "S", "L", "ms", "U", "us", "N"]))
    mult = draw(integers(min_value=1, max_value=10))
    value = f"{mult}{freq}"
    return value


@composite
def dataFieldFlags(draw):
    data = draw(dioses())
    field = draw(sampled_from(sorted(data.columns)))
    flags = draw(flagses(data))
    return data, field, flags


@composite
def functionCalls(draw, module: str = None):
    func = draw(functions(module))
    kwargs = draw(functionKwargs(func))
    return func, kwargs


@contextmanager
def applyStrategies(strategies: dict):

    for dtype, strategy in strategies.items():
        register_type_strategy(dtype, strategy)

    yield

    for dtype in strategies.keys():
        del _global_type_lookup[dtype]


@composite
def functionKwargs(draw, func):
    data = draw(dioses())
    field = draw(sampled_from(sorted(data.columns)))

    kwargs = {"data": data, "field": field, "flags": draw(flagses(data))}

    i64 = np.iinfo("int64")

    strategies = {
        FreqString: frequencyStrings,
        ColumnName: lambda _: sampled_from(
            sorted(c for c in data.columns if c != field)
        ),
        IntegerWindow: lambda _: integers(min_value=1, max_value=len(data[field]) - 1),
        int: lambda _: integers(min_value=i64.min + 1, max_value=i64.max - 1),
    }

    with applyStrategies(strategies):
        for k, v in get_type_hints(func).items():
            if k not in {"data", "field", "flags", "return"}:
                value = draw(from_type(v))
                kwargs[k] = value

    return kwargs
